r"""
fgout_tools module: $CLAW/geoclaw/src/python/geoclaw/fgout_tools.py

Tools to specify and work with fgout grids, used to output GeoClaw solutions
on a fixed grid at a sequence of times, regardless of the AMR structure.

Includes:

- class FGoutFrame: used to hold a single frame of fgout output data
- class FGoutGrid: used to specify and store info about an fgout grid, with
                   methods to read and write info to fgout_grids.data
- function make_fgout_fcn_xy: Takes an FGoutFrame object and produces an
            interpolating function that can be evaluated for any (x,y).
- function make_fgout_fcn_xyt: Takes 2 FGoutFrame objects and produces an
            interpolating function that can be evaluated for any (x,y,t)
            at intermediate times.
- function write_netcdf: Write a specified set of qoi's from a list of
            fgout frames, as a single netCDF file
- function read_netcdf: Read a netCDF file and return a list of fgout frames,
            assuming the file contains all the qoi's needed to reconstruct q.
- function read_netcdf_arrays: Read a netCDF file and extract the
            requested quantities of interest as numpy arrays.
- print_netcdf_info: Print info about the contents of a netCDF file containing
            some fgout frames.
"""

import os
from numpy import sqrt, ma, mod
import numpy

class FGoutFrame(object):

    """
    Class to hold a single frame of fgout data at one output time.
    Several attributes are defined as properties that can be evaluated
    and stored only when needed by the user.
    """

    def __init__(self, fgout_grid, frameno=None):
        self.fgout_grid = fgout_grid
        self.frameno = frameno
        self.t = None

        # private attributes for those that are only created if
        # needed by the user:
        self._h = None
        self._hu = None
        self._hv = None
        self._eta = None
        self._B = None
        self._u = None
        self._v = None
        self._s = None
        self._hss = None
        self._hm = None

    # Define shortcuts to attributes of self.fgout_grid that are the same
    # for all frames (e.g. X,Y) to avoid storing grid for every frame.

    @property
    def x(self):
        return self.fgout_grid.x

    @property
    def y(self):
        return self.fgout_grid.y

    @property
    def X(self):
        return self.fgout_grid.X

    @property
    def Y(self):
        return self.fgout_grid.Y

    @property
    def delta(self):
        return self.fgout_grid.delta

    @property
    def extent_centers(self):
        return self.fgout_grid.extent_centers

    @property
    def extent_edges(self):
        return self.fgout_grid.extent_edges

    @property
    def drytol(self):
        return self.fgout_grid.drytol

    @property
    def qmap(self):
        return self.fgout_grid.qmap

    # Define attributes such as h as @properties with lazy evaluation:
    # the corresponding array is created and stored only when first
    # accessed by the user.  Those not needed are not created.

    @property
    def h(self):
        """depth"""
        if self._h is None:
            q_out_vars = self.fgout_grid.q_out_vars
            try:
                i_h = q_out_vars.index(self.qmap['h'])
                self._h = self.q[i_h,:,:]
            except ValueError: # if q_out_vars does not contain qmap['h'], a value error is thrown
                try:
                    i_eta = q_out_vars.index(self.qmap['eta'])
                    i_B = q_out_vars.index(self.qmap['B'])
                    self._h = self.q[i_eta,:,:] - self.q[i_B,:,:]
                except ValueError:
                    print('*** Could not find h or eta-B in q_out_vars')
                    raise AttributeError
        return self._h

    @property
    def hu(self):
        """momentum h*u"""
        if self._hu is None:
            q_out_vars = self.fgout_grid.q_out_vars
            try:
                i_hu = q_out_vars.index(self.qmap['hu'])
                self._hu = self.q[i_hu,:,:]
            except:
                print('*** Could not find hu in q_out_vars')
                raise
        return self._hu

    @property
    def u(self):
        """speed u, computed as hu/h or set to 0 if h<self.drytol"""
        if self._u is None:
            self._u = numpy.divide(self.hu, self.h,\
                              out=numpy.zeros(self.h.shape, dtype=float), \
                              where=(self.h>self.drytol))
        return self._u

    @property
    def hv(self):
        """momentum h*v"""
        if self._hv is None:
            q_out_vars = self.fgout_grid.q_out_vars
            try:
                i_hv = q_out_vars.index(self.qmap['hv'])
                self._hv = self.q[i_hv,:,:]
            except:
                print('*** Could not find hv in q_out_vars')
                raise
        return self._hv

    @property
    def v(self):
        """speed v, computed as hv/h or set to 0 if h<self.drytol"""
        if self._v is None:
            self._v = numpy.divide(self.hv, self.h,\
                              out=numpy.zeros(self.h.shape, dtype=float), \
                              where=(self.h>self.drytol))
        return self._v

    @property
    def eta(self):
        """surface eta = h+B"""
        if self._eta is None:
            q_out_vars = self.fgout_grid.q_out_vars
            try:
                i_eta = q_out_vars.index(self.qmap['eta'])
                self._eta = self.q[i_eta,:,:]
                #print('+++qmap["eta"] = %i' % self.qmap["eta"])
                #print('+++i_eta = %i' % i_eta)
            except:
                try:
                    i_h = q_out_vars.index(self.qmap['h'])
                    i_B = q_out_vars.index(self.qmap['B'])
                    self._eta = self.q[i_h,:,:] + self.q[i_B,:,:]
                except:
                    print('*** Could not find eta or h+B in q_out_vars')
                    raise
        return self._eta

    @property
    def B(self):
        """topography"""
        if self._B is None:
            q_out_vars = self.fgout_grid.q_out_vars
            try:
                i_B = q_out_vars.index(self.qmap['B'])
                self._B = self.q[i_B,:,:]
                #print('+++qmap["B"] = %i' % self.qmap["B"])
                #print('+++i_B = %i' % i_B)
            except:
                try:
                    i_h = q_out_vars.index(self.qmap['h'])
                    i_eta = q_out_vars.index(self.qmap['eta'])
                    self._B = self.q[i_eta,:,:] - self.q[i_h,:,:]
                    #print('+++ computing B: i_h = %i, i_eta = %i' % (i_h,i_eta))
                    #print('+++qmap["h"] = %i' % self.qmap["h"])
                    #print('+++qmap["eta"] = %i' % self.qmap["eta"])
                except:
                    print('*** Could not find B or eta-h in q_out_vars')
                    raise
        return self._B

    @property
    def s(self):
        """speed s = sqrt(u**2 + v**2)"""
        if self._s is None:
            self._s = numpy.sqrt(self.u**2 + self.v**2)
        return self._s

    @property
    def hss(self):
        """momentum flux h*s**2"""
        if self._hss is None:
            self._hss = self.h * self.s**2
        return self._hss

    @property
    def huc(self):
        """huc - Boussinesq correction to hu"""
        if self._huc is None:
            q_out_vars = self.fgout_grid.q_out_vars
            try:
                i_huc = q_out_vars.index(self.qmap['huc'])
                self._huc = self.q[i_huc,:,:]
            except:
                print('*** Could not find huc in q_out_vars')
                raise
        return self._huc

    @property
    def hvc(self):
        """hvc - Boussinesq correction to hv"""
        if self._hvc is None:
            q_out_vars = self.fgout_grid.q_out_vars
            try:
                i_hvc = q_out_vars.index(self.qmap['hvc'])
                self._hvc = self.q[i_hvc,:,:]
            except:
                print('*** Could not find hvc in q_out_vars')
                raise
        return self._hvc

    @property
    def hm(self):
        """dclaw: h * mass fraction"""
        if self._hm is None:
            q_out_vars = self.fgout_grid.q_out_vars
            try:
                i_hm = q_out_vars.index(self.qmap['hm'])
                self._hm = self.q[i_hm,:,:]
                #print('+++qmap["hm"] = %i' % self.qmap["hm"])
                #print('+++i_hm = %i' % i_hm)
            except:
                print('*** Could not find hm in q_out_vars')
                raise
        return self._hm

    @property
    def pb(self):
        """dclaw variable """
        if self._pb is None:
            q_out_vars = self.fgout_grid.q_out_vars
            try:
                i_pb = q_out_vars.index(self.qmap['pb'])
                self._pb = self.q[i_pb,:,:]
            except:
                print('*** Could not find pb in q_out_vars')
                raise
        return self._pb

    @property
    def hchi(self):
        """dclaw variable """
        if self._hchi is None:
            q_out_vars = self.fgout_grid.q_out_vars
            try:
                i_hchi = q_out_vars.index(self.qmap['hchi'])
                self._hchi = self.q[i_hchi,:,:]
            except:
                print('*** Could not find hchi in q_out_vars')
                raise
        return self._hchi

    @property
    def bdif(self):
        """dclaw variable """
        if self._bdif is None:
            q_out_vars = self.fgout_grid.q_out_vars
            try:
                i_bdif = q_out_vars.index(self.qmap['bdif'])
                self._bdif = self.q[i_bdif,:,:]
            except:
                print('*** Could not find bdif in q_out_vars')
                raise
        return self._bdif


class FGoutGrid(object):

    """
    New class introduced in 5.9.0 to keep store information both about the
    fgout input data and the output generated by a GeoClaw run.
    """

    def __init__(self,fgno=None,outdir='.',output_format=None,
                 qmap='geoclaw'):


        # mapping from variable names to possible values in q_out_vars
        if type(qmap) is dict:
            self.qmap = qmap
        elif qmap == 'geoclaw':
            # default for GeoClaw:
            self.qmap = {'h':1, 'hu':2, 'hv':3, 'eta':4, 'B':5}
        elif qmap == 'geoclaw-bouss':
            self.qmap = {'h':1, 'hu':2, 'hv':3, 'huc':4, 'hvc':5,
                         'eta':6, 'B':7}
        elif qmap == 'dclaw':
            self.qmap = {'h':1, 'hu':2, 'hv':3, 'hm':4, 'pb':5, 'hchi':6,
                         'bdif':7, 'eta':8, 'B':9}
        else:
            raise ValueError('Invalid qmap: %s' % qmap)

        # GeoClaw input values:
        # Many of these should be read from fgout_grids.data
        # using read_fgout_grids_data before using read_frame
        self.id = ''  # identifier, optional
        self.point_style = 2  # only option currently supported
        self.npts = None
        self.nx = None
        self.ny = None
        self.output_style = 1
        self.tstart =  None
        self.tend = None
        self.nout = None
        self.fgno = fgno
        self.outdir = outdir

        # Note output_format will be reset by read_fgout_grids_data:
        self.output_format = output_format

        self.q_out_vars = [1,2,3,4]  # list of which components to print
                                          # default: h,hu,hv,eta (5=topo)
        self.nqout = None # number of vars to print out

        self.drytol = 1e-3          # used for computing u,v from hu,hv

        # private attributes for those that are only created if
        # needed by the user:

        self._X = None
        self._Y = None
        self._x = None
        self._y = None
        self._delta = None           # (dx,dy)
        self._extent_centers = None  # defined by fgout points
        self._extent_edges = None    # extended so points are cell centers

        self._plotdata = None

    @property
    def x(self):
        """1D array x of longitudes (cell centers)"""
        if self._x is None:
            dx = (self.x2 - self.x1)/self.nx
            self._x = numpy.linspace(self.x1+dx/2, self.x2-dx/2, self.nx)
        return self._x

    @property
    def y(self):
        """1D array y of latitudes (cell centers)"""
        if self._y is None:
            dy = (self.y2 - self.y1)/self.ny
            self._y = numpy.linspace(self.y1+dy/2, self.y2-dy/2, self.ny)
        return self._y

    @property
    def X(self):
        """2D array X of longitudes (cell centers)"""
        if self._X is None:
            self._X = numpy.meshgrid(self.x, self.y)[0].T
        return self._X

    @property
    def Y(self):
        """2D array Y of latitudes (cell centers)"""
        if self._Y is None:
            self._Y = numpy.meshgrid(self.x, self.y)[1].T
        return self._Y

    @property
    def delta(self):
        if self._delta is None:
            dx = (self.x2 - self.x1)/self.nx
            dy = (self.y2 - self.y1)/self.ny
            self._delta = (dx,dy)
        return self._delta

    @property
    def extent_centers(self):
        """Extent of cell centers [xmin,xmax,ymin,ymax]"""
        if self._extent_centers is None:
            self._extent_centers = [self.x.min(), self.x.max(),\
                                    self.y.min(), self.y.max()]
        return self._extent_centers

    @property
    def extent_edges(self):
        """Extent of cell edges [xmin,xmax,ymin,ymax]"""
        if self._extent_edges is None:
            dx,dy = self.delta
            self._extent_edges = [self.x.min()-dx/2, self.x.max()+dx/2,\
                                    self.y.min()-dy/2, self.y.max()+dy/2]
        return self._extent_edges


    # Create plotdata of class clawpack.visclaw.ClawPlotData
    # only when needed for reading GeoClaw output,
    # since this must be done after fgno, outdir, output_format
    # have been specified:

    @property
    def plotdata(self):
        if self._plotdata is None:
            self._plotdata = self.set_plotdata()
        return self._plotdata

    def set_plotdata(self):
        """
        Create a plotdata, assuming attributes fgno, outdir, output_format
        have all been set.
        """
        from clawpack.visclaw.data import ClawPlotData

        assert self.fgno is not None, '*** fgno must be set'
        assert self.outdir is not None, '*** outdir must be set'
        assert self.output_format is not None, '*** output_format must be set'
        if 0:
            print('+++ creating plotdata for fgno=%i, outdir=%s, format=%s' \
                  % (self.fgno,self.outdir,self.output_format))

        plotdata = ClawPlotData()
        plotdata.outdir = self.outdir
        plotdata.format = self.output_format
        plotdata.file_prefix = 'fgout%s' % str(self.fgno).zfill(4)
        if self.output_format[:6]=='binary':
            # could be 'binary', 'binary32' or 'binary64'
            file_prefix_str = plotdata.file_prefix + '.b'
        else:
            file_prefix_str = plotdata.file_prefix + '.q'

        # Contrary to usual default, set save_frames to False so that all
        # the fgout frames read in are not saved in plotdata.framesoln_dict.
        # Otherwise this might be a memory hog when making an animation with
        # many frames on a fine fgout grid.
        plotdata.save_frames = False

        return plotdata


    def read_fgout_grids_data(self, fgno=None, data_file='fgout_grids.data'):
        """
        Read input info for fgout grid number fgno from the data file
        fgout_grids.data, which should have been created by setrun.py.
        This file now contains info about all fgout grids.
        """
        if fgno is not None:
            self.fgno = fgno
        assert self.fgno is not None, '*** fgno must be set'

        data_path = os.path.join(self.outdir, data_file)
        print('Reading fgout grid info from \n    %s' % data_path)

        with open(data_path) as filep:
            lines = filep.readlines()
        fgout_input = None
        for lineno,line in enumerate(lines):
            if 'fgno' in line:
                if int(line.split()[0]) == self.fgno:
                    fgout_input = lines[lineno+1:]
                    #print('Found line %i: %s' % (lineno,line))
                    break

        if fgout_input is None:
            raise ValueError('fgout grid fgno = %i not found in %s' \
                             % (self.fgno, data_file))

        lineno = 0 # next input line
        next_value = fgout_input[lineno].split()[lineno]  # a string
        next_value_int = ('.' not in next_value)  # should be True in v5.11
        err_msg = '\n*** Expecting integer output_style next in %s' \
            % data_file \
            + '\n    If this is fgout data from before v5.11, try using' \
            + '\n    read_fgout_grids_data_pre511'
        assert next_value_int, err_msg
        self.output_style = int(next_value)
        lineno += 1
        if (self.output_style == 1):
            # equally spaced times:
            self.nout = int(fgout_input[lineno].split()[0])
            lineno += 1
            self.tstart = float(fgout_input[lineno].split()[0])
            lineno += 1
            self.tend = float(fgout_input[lineno].split()[0])
            lineno += 1
            self.times = numpy.linspace(self.tstart, self.tend, self.nout)
        elif (self.output_style == 2):
            # list of times:
            self.nout = int(fgout_input[lineno].split()[0])
            lineno += 1
            times_str = fgout_input[lineno].split()[:self.nout]
            self.times = numpy.array([float(ts) for ts in times_str])
            lineno += 1
        else:
            raise ValueError('Unrecognized fgout output_style: %s' \
                             % self.output_style)
        self.point_style = point_style = int(fgout_input[lineno].split()[0])
        lineno += 1
        output_format = int(fgout_input[lineno].split()[0])
        lineno += 1
        if output_format == 1:
            self.output_format = 'ascii'
        elif output_format == 2:
            self.output_format = 'binary32'
        elif output_format == 3:
            self.output_format = 'binary'
        else:
            raise NotImplementedError("fgout not implemented for " \
                    + "output_format %i"  % output_format)
        print('Reading input for fgno=%i, point_style = %i ' \
                % (self.fgno, self.point_style))
        if point_style == 2:
            self.nx = nx = int(fgout_input[lineno].split()[0])
            self.ny = ny = int(fgout_input[lineno].split()[1])
            lineno += 1
            self.x1 = float(fgout_input[lineno].split()[0])
            self.y1 = float(fgout_input[lineno].split()[1])
            lineno += 1
            self.x2 = float(fgout_input[lineno].split()[0])
            self.y2 = float(fgout_input[lineno].split()[1])
            lineno += 1
        else:
            raise NotImplementedError("fgout not implemented for point_style %i" \
                % point_style)
        tokens = fgout_input[lineno].split()
        self.q_out_vars = []
        for token in tokens:
            try:
                self.q_out_vars.append(int(token))
            except:
                break
        print('Found fgout grid q_out_vars = ',self.q_out_vars)
        print('Using this mapping to fgout variable names: ')
        print('      qmap = ',self.qmap)

    def read_fgout_grids_data_pre511(self, fgno=None,
                                     data_file='fgout_grids.data'):
        """
        For backward compatibility, this reads fgout_grids.data files
        in the format used prior to v5.11.

        In this case, the following values are used, as set in __init__():
            self.output_style = 1
            self.qmap = 'qmap'
            self.q_out_vars = [1,2,3,4]  # h,hu,hv,eta

        Read input info for fgout grid number fgno from the data file
        fgout_grids.data, which should have been created by setrun.py.
        This file now contains info about all fgout grids.
        """

        if fgno is not None:
            self.fgno = fgno
        assert self.fgno is not None, '*** fgno must be set'

        data_path = os.path.join(self.outdir, data_file)
        print('Reading fgout grid info from \n    %s' % data_path)

        with open(data_path) as filep:
            lines = filep.readlines()
        fgout_input = None
        for lineno,line in enumerate(lines):
            if 'fgno' in line:
                if int(line.split()[0]) == self.fgno:
                    fgout_input = lines[lineno+1:]
                    #print('Found line %i: %s' % (lineno,line))
                    break

        if fgout_input is None:
            raise ValueError('fgout grid fgno = %i not found in %s' \
                             % (fgno, data_file))

        self.tstart = float(fgout_input[0].split()[0])
        self.tend = float(fgout_input[1].split()[0])
        self.nout = int(fgout_input[2].split()[0])
        self.point_style = point_style = int(fgout_input[3].split()[0])
        output_format = int(fgout_input[4].split()[0])
        if output_format == 1:
            self.output_format = 'ascii'
        elif output_format == 3:
            self.output_format = 'binary'
        print('Reading input for fgno=%i, point_style = %i ' \
                % (self.fgno, self.point_style))
        if point_style == 2:
            self.nx = nx = int(fgout_input[5].split()[0])
            self.ny = ny = int(fgout_input[5].split()[1])
            self.x1 = float(fgout_input[6].split()[0])
            self.y1 = float(fgout_input[6].split()[1])
            self.x2 = float(fgout_input[7].split()[0])
            self.y2 = float(fgout_input[7].split()[1])
        else:
            raise NotImplementedError("fgout not implemented for point_style %i" \
                % point_style)

    def write_to_fgout_data(self, fid):
        """
        Convert fgout data specified in setrun.py to file `fgout_grids.data`
        read in by GeoClaw fortran code.
        """

        print("\n---------------------------------------------- ")
        assert self.point_style is not None, 'Need to set point_style'

        point_style = self.point_style
        if point_style not in [2]:
            errmsg = 'point_style %s is not implement, only point_style==2' \
                                    % point_style
            raise NotImplementedError(errmsg)

        if self.output_format == 'ascii':
            output_format = 1
        elif self.output_format == 'binary32':
            output_format = 2
        elif self.output_format in ['binary','binary64']:
            output_format = 3
        else:
            errmsg = "fgout output_format must be ascii, binary32, or binary64"
            raise NotImplementedError(errmsg)



        # write header, independent of point_style:
        #fid = open(self.input_file_name,'w')
        fid.write("\n")
        fid.write("%i                           # fgno\n" % self.fgno)

        fid.write("%i                           # output_style\n" \
                    % self.output_style)

        if self.output_style == 1:
            assert self.tstart is not None, 'Need to set tstart'
            assert self.tend is not None, 'Need to set tend'
            assert self.nout is not None, 'Need to set nout'
            fid.write("%i %s           # nout\n" % (self.nout, 11*" "))
            fid.write("%16.10e            # tstart\n"  % self.tstart)
            fid.write("%16.10e            # tend\n"  % self.tend)
        elif self.output_style == 2:
            self.nout = len(self.output_times)
            fid.write("%i %s           # nout\n" % (self.nout, 11*" "))

            # remove [] and , from list of times:
            output_times_str = repr(list(self.output_times))[1:-1]
            output_times_str = output_times_str.replace(',','')
            fid.write("%s            # output_times\n"  % output_times_str)
        else:
            raise ValueError('fgout output_style must be 1 or 2')
        fid.write("%i %s              # point_style\n" \
                            % (self.point_style,12*" "))
        fid.write("%i %s              # output_format\n" \
                            % (output_format,12*" "))

        print('fgout grid %i has point_style = %i' % (self.fgno, point_style))


        if point_style == 2:
            # 2d grid of points
            x1,x2 = self.x1, self.x2
            y1,y2 = self.y1, self.y2
            nx,ny = self.nx, self.ny

            dx = (x2-x1)/nx   # x1,x2 are cell edges
            dy = (y2-y1)/ny   # y1,y2 are cell edges

            npts = nx*ny

            fid.write("%i  %i %s          # nx,ny\n" \
                                % (nx,ny,10*" "))
            fid.write("%16.10e   %20.10e            # x1, y1\n" % (x1,y1))
            fid.write("%16.10e   %20.10e            # x2, y2\n" % (x2,y2))

            print("   specifying fgout grid with shape %i by %i, with  %i points" \
                    % (nx,ny,npts))
            print("   lower left  = (%15.10f,%15.10f)" % (x1,y1))
            print("   upper right = (%15.10f,%15.10f)" % (x2,y2))
            print("   dx = %15.10e,  dy = %15.10e" % (dx,dy))

        # q_out_vars is a list of q components to print, e.g. [1,4]

        format = len(self.q_out_vars) * '%s '
        fid.write(format % tuple(self.q_out_vars)+ " # q_out_vars\n")
        fid.write('\n')


    def read_frame(self, frameno):
        """
        Read a single frame of fgout data.
        """

        from datetime import timedelta

        try:
            fgoutX = self.X
            fgoutY = self.Y
        except:
            msg = '\n*** Before reading frame, you must set FGoutGrid data,' \
                  '\n*** Typically by calling read_fgout_grids_data'
            raise ValueError(msg)

            # prior to v5.11, self.read_fgout_grids_data() called here
            # rather than raising exception...
            print(msg)
            print('*** Calling read_fgout_grids_data...')
            self.read_fgout_grids_data()
            fgoutX = self.X
            fgoutY = self.Y

        try:
            fr = self.plotdata.getframe(frameno)
        except:
            print('*** Could not read fgout grid %i frame %i from %s' \
                 % (self.fgno,frameno,self.plotdata.outdir))
            raise
        state = fr.states[0]  # only 1 AMR grid
        patch = state.patch

        fgout_frame = FGoutFrame(self, frameno)
        fgout_frame.fgout_grid = self

        fgout_frame.q = state.q

        fgout_frame.t = state.t

        fgout_frame.frameno = frameno

        X,Y = patch.grid.p_centers[:2]

        if not numpy.allclose(X, fgoutX):
            errmsg = '*** X read from output does not match fgout_grid.X'
            raise ValueError(errmsg)

        if not numpy.allclose(Y, fgoutY):
            errmsg = '*** Y read from output does not match fgout_grid.Y'
            raise ValueError(errmsg)

        return fgout_frame


# ========================
# Functions for interpolating from fgout grid to arbitrary points,
# useful for example if using velocity field to model particle/debris motion

def make_fgout_fcn_xy(fgout, qoi, method='nearest',
                       bounds_error=False, fill_value=numpy.nan):
    """
    Create a function that can be called at (x,y) and return the qoi
    interpolated in space from the fgout array.

    qoi should be a string (e.g. 'h', 'u' or 'v') corresponding to
    an attribute of fgout.

    The function returned takes arguments x,y that can be floats or
    (equal length) 1D arrays of values that lie within the spatial
    extent of fgout.

    bounds_error and fill_value determine the behavior if (x,y) is not in
    the bounds of the data, as in scipy.interpolate.RegularGridInterpolator.
    """

    from scipy.interpolate import RegularGridInterpolator

    try:
        q = getattr(fgout,qoi)
    except:
        print('*** fgout missing attribute qoi = %s?' % qoi)

    err_msg = '*** q must have same shape as fgout.X\n' \
            + 'fgout.X.shape = %s,   q.shape = %s' % (fgout.X.shape,q.shape)
    assert fgout.X.shape == q.shape, err_msg

    x1 = fgout.X[:,0]
    y1 = fgout.Y[0,:]
    fgout_fcn1 = RegularGridInterpolator((x1,y1), q, method=method,
                bounds_error=bounds_error, fill_value=fill_value)

    def fgout_fcn(x,y):
        """
        Function that can be evaluated at single point or arrays (x,y).
        """
        from numpy import array, vstack
        xa = array(x)
        ya = array(y)
        xyout = vstack((xa,ya)).T
        qout = fgout_fcn1(xyout)
        if len(qout) == 1:
            qout = qout[0]  # return scalar
        return qout

    return fgout_fcn


def make_fgout_fcn_xyt(fgout1, fgout2, qoi, method_xy='nearest',
                       method_t='linear', bounds_error=False,
                       fill_value=numpy.nan):
    """
    Create a function that can be called at (x,y,t) and return the qoi
    interpolated in space and time between the two frames fgout1 and fgout2.

    qoi should be a string (e.g. 'h', 'u' or 'v') corresponding to
    an attribute of fgout.

    method_xy is the method used in creating the spatial interpolator,
    and is passed to make_fgout_fcn_xy.

    method_t is the method used for interpolation in time, currently only
    'linear' is supported, which linearly interpolates.

    bounds_error and fill_value determine the behavior if (x,y,t) is not in
    the bounds of the data.

    The function returned takes arguments x,y (floats or equal-length 1D arrays)
    of values that lie within the spatial extent of fgout1, fgout2
    (which are assumed to cover the same uniform grid at different times)
    and t should be a float that lies between fgout1.t and fgout2.t.
    """

    assert numpy.allclose(fgout1.X, fgout2.X), \
                            '*** fgout1 and fgout2 must have same X'
    assert numpy.allclose(fgout1.Y, fgout2.Y), \
                            '*** fgout1 and fgout2 must have same Y'

    t1 = fgout1.t
    t2 = fgout2.t
    #assert t1 < t2, '*** expected fgout1.t < fgout2.t'

    fgout1_fcn_xy = make_fgout_fcn_xy(fgout1, qoi, method=method_xy,
                       bounds_error=bounds_error, fill_value=fill_value)
    fgout2_fcn_xy = make_fgout_fcn_xy(fgout2, qoi, method=method_xy,
                       bounds_error=bounds_error, fill_value=fill_value)

    def fgout_fcn(x,y,t):
        """
        Function that can be evaluated at single point or arrays (x,y)
        at a single time t.
        """
        from numpy import array, ones
        xa = array(x)
        ya = array(y)
        tol = 1e-6  # to make sure it works ok when called with t=t1 or t=t2
        if t1-tol <= t <= t2+tol:
            alpha = (t-t1)/(t2-t1)
        elif bounds_error:
            errmsg = '*** argument t=%g should be between t1=%g and t2=%g' \
                     % (t,t1,t2)
            raise ValueError(errmsg)
        else:
            qout = fill_value * ones(xa.shape)
            return qout

        qout1 = fgout1_fcn_xy(x,y)
        qout2 = fgout2_fcn_xy(x,y)

        if method_t == 'linear':
            if t1 <= t <= t2:
                alpha = (t-t1)/(t2-t1)
                qout = (1-alpha)*qout1 + alpha*qout2
        else:
            raise NotImplementedError('method_t = %s not supported' % method_t)

        return qout

    return fgout_fcn

# ===============================
# Functions for writing a set of fgout frames as a netCDF file, and
# reading such a file:

def write_netcdf(fgout_frames, fname_nc='fgout_frames.nc',
                 qois = ['h','hu','hv','eta'], datatype='f4',
                 include_B0=False, include_Bfinal=False,
                 description='', verbose=True):
    """
    Write a list of fgout frames (at different times on the same rectangular
    grid) to a single netCDF file, with some metadata and the topography,
    if desired.

    fgout_frames should be a list of FGoutFrame objects, all of the same size
    and at increasing times.

    fname_nc is the name of the file to write.

    qois is a list of strings, the quantities of interest to include in file.
        This could include any of:

            'h', 'eta', 'hu', 'hv', 'u', 'v', 's', 'hss', 'B'.

        All other quantities can be computed from h, hu, hv, eta,
        the original fgout variables from GeoClaw, but for some applications
        you might only want to save 'h' and 's', for example.

    datatype should be 'f4' [default] or 'f8', specifying bytes per qoi value.
        'f8' has full precision of the original data, but the file will be
        twice as large and may not be needed for downstream applications.

    Note that the topography B = eta - h, so it is not necessary to store all
    three of these.  Also, B is often the same for all frames, so rather than
    storing B at each frame as a qoi, two other options are also provided
    (and then storing eta or h for all frames allows calculating the other):

    include_Bfinal: If True, include the topography B array from the final frame
    as the Bfinal array.

    include_B0: If True, include the topography B array from the first frame
    as the B0 array.  This is only useful if, e.g., the first frame is initial
    topography before co-seismic deformation, and at later times the topography
    is always equal to Bfinal.

    `description` is a string that will be added as metadata.
    A metadata field `history` will also be added, which includes the
    time the file was created and the path to the directory where it was made.
    """

    import netCDF4
    from datetime import datetime, timedelta
    from cftime import num2date, date2num
    import time
    timestr = time.ctime(time.time())  # current time for metadata

    fg_times = numpy.array([fg.t for fg in fgout_frames])

    if verbose:
        print('Creating %s with fgout frames at times: ' % fname_nc)
        print(fg_times)

    fg0 = fgout_frames[0]
    x = fg0.x
    y = fg0.y

    xs = numpy.array([fg.x for fg in fgout_frames])
    ys = numpy.array([fg.y for fg in fgout_frames])
    # assert same for all times

    units = {'h':'meters', 'eta':'meters', 'hu':'m^2/s', 'hv':'m^2/s',
             'u':'m/s', 'v':'m/s', 's':'m/s', 'hss':'m^3/s^2', 'B':'meters'}

    with netCDF4.Dataset(fname_nc, 'w') as rootgrp:

        rootgrp.description = description
        rootgrp.history = "Created " + timestr
        rootgrp.history += " in %s;  " % os.getcwd()

        lon = rootgrp.createDimension('lon', len(x))
        longitudes = rootgrp.createVariable('lon','f8',('lon',))
        longitudes[:] = x
        longitudes.units = 'degrees_east'

        lat = rootgrp.createDimension('lat', len(y))
        latitudes = rootgrp.createVariable('lat','f8',('lat',))
        latitudes[:] = y
        latitudes.units = 'degrees_north'

        time = rootgrp.createDimension('time', len(fg_times))
        times = rootgrp.createVariable('time','f8',('time',))
        times[:] = fg_times
        times.units = 'seconds'

        if 0:
            # Could make times be datetimes relative to some event time, e.g.:
            times.units = 'seconds since 1700-01-26 21:00:00.0'
            times.calendar = 'gregorian'
            dates = [datetime(1700,1,26,21) + timedelta(seconds=ss) \
                        for ss in fg_times]
            times[:] = date2num(dates,units=times.units,calendar=times.calendar)

        if include_B0:
            B0 = rootgrp.createVariable('B0',datatype,('lon','lat',))
            B0[:,:] = fg0.B
            B0.units = 'meters'

        if include_Bfinal:
            fg_final = fgout_frames[-1]
            Bfinal = rootgrp.createVariable('Bfinal',datatype,('lon','lat',))
            Bfinal[:,:] = fg_final.B
            Bfinal.units = 'meters'

        for qoi in qois:
            qoi_frames = [getattr(fgout,qoi) for fgout in fgout_frames]
            qoi_var = rootgrp.createVariable(qoi,datatype,('time','lon','lat',))
            qoi_var[:,:,:] = qoi_frames
            qoi_var.units = units[qoi]

def get_as_array(var, rootgrp, verbose=True):
    """
    Utility function to retrieve variable from netCDF file and convert to
    numpy array.
    """
    a = rootgrp.variables.get(var, None)
    if a is not None:
        if verbose: print('    Loaded %s with shape %s' % (var,repr(a.shape)))
        return numpy.array(a)
    else:
        if verbose: print('    Did not find %s' % var)
        return None


def read_netcdf_arrays(fname_nc, qois, verbose=True):
    """
    Read a netCDF file and extract the quantities of interest denoted by
    strings in the list qois, which can include:

        'h', 'eta', 'hu', 'hv', 'u', 'v', 's', 'hss', 'B'.

    qois can also include the time-independent 'B0' and/or 'Bfinal'

    Returns

        x, y, t, qoi_arrays

    where x,y define the longitude, latitudes, t is the times of the frames,
    and qoi_arrays is a dictionary indexed by the strings from qois.

    Each dict element is an array with shape (len(t), len(x), len(y))
    for time-dependent qoi's, or (len(x), len(y)) for B0 or Bfinal,
    or None if that qoi was not found in the netCDF file.
    """

    import netCDF4

    with netCDF4.Dataset(fname_nc, 'r') as rootgrp:
        if verbose:
            print('Reading data to fgout frames from nc file',fname_nc)
            print('        nc file description: ', rootgrp.description)
            print('History:  ', rootgrp.history)

        x = get_as_array('lon', rootgrp, verbose)
        y = get_as_array('lat', rootgrp, verbose)
        t = get_as_array('time', rootgrp, verbose)

        vars = list(rootgrp.variables)

        qoi_arrays = {}
        for qoi in qois:
            qoi_array = get_as_array(qoi, rootgrp, verbose)
            qoi_arrays[qoi] = qoi_array

    return x,y,t,qoi_arrays



def read_netcdf(fname_nc, fgout_grid=None, verbose=True):
    """
    Read a netCDF file and return a list of FGoutFrame instances.
    This will only be possible if the netCDF file contains at least
    the qoi's 'h','hu','hv','eta' required to reconstruct the q array
    as output by GeoClaw.
    """

    import netCDF4

    with netCDF4.Dataset(fname_nc, 'r') as rootgrp:
        if 1:
            print('Reading data to fgout frames from nc file',fname_nc)
            print('        nc file description: ', rootgrp.description)
            print('History:  ', rootgrp.history)

        x = get_as_array('lon', rootgrp, verbose)
        y = get_as_array('lat', rootgrp, verbose)
        t = get_as_array('time', rootgrp, verbose)

        vars = list(rootgrp.variables)


        for qoi in ['h','hu','hv','eta']:
            errmsg = '*** Cannot reconstruct fgout frame without %s' % qoi
            assert qoi in vars, errmsg
        h = get_as_array('h', rootgrp, verbose)
        hu = get_as_array('hu', rootgrp, verbose)
        hv = get_as_array('hv', rootgrp, verbose)
        eta = get_as_array('eta', rootgrp, verbose)

        if (x is None) or (y is None):
            print('*** Could not create grid')
        else:
            X,Y = numpy.meshgrid(x,y)

        fgout_frames = []

        for k in range(eta.shape[0]):
            fgout = FGoutFrame(fgout_grid=fgout_grid, frameno=k)
            fgout.x = x
            fgout.y = y
            fgout.t = t[k]
            fgout.q = numpy.empty((4,eta.shape[1],eta.shape[2]))
            fgout.q[0,:,:] = h[k,:,:]
            fgout.q[1,:,:] = hu[k,:,:]
            fgout.q[2,:,:] = hv[k,:,:]
            fgout.q[3,:,:] = eta[k,:,:]
            fgout.X = X
            fgout.Y = Y
            fgout_frames.append(fgout)

        print('Created fgout_frames as list of length %i' % len(fgout_frames))

    return fgout_frames

def print_netcdf_info(fname_nc):
    """
    Print out info about the contents of a netCDF file contining fgout frames,
    written using write_netcdf.
    """
    import netCDF4

    with netCDF4.Dataset(fname_nc, 'r') as rootgrp:
        x = get_as_array('lon', rootgrp, verbose=False)
        y = get_as_array('lat', rootgrp, verbose=False)
        t = get_as_array('time', rootgrp, verbose=False)

        vars = list(rootgrp.variables)

        print('===================================================')
        print('netCDF file %s contains:' % fname_nc)
        print('description: \n', rootgrp.description)
        print('history: \n', rootgrp.history)
        print('%i longitudes from %.6f to %.6f' % (len(x),x[0],x[-1]))
        print('%i latitudes from %.6f to %.6f' % (len(y),y[0],y[-1]))
        print('%i times from %.3f to %.3f' % (len(t),t[0],t[-1]))
        print('variables: ',vars)
        print('===================================================')




""" 
Set up the plot figures, axes, and items to be done for each frame.

This module is imported by the plotting routines and then the
function setplot is called to set the plot parameters.
    
""" 

from __future__ import absolute_import
from __future__ import print_function
import os
if os.path.exists('./1drad/_output'):
    qref_dir = os.path.abspath('./1drad/_output')
else:
    qref_dir = None
    print("Directory ./1drad/_output not found")


#--------------------------
def setplot(plotdata=None):
#--------------------------
    
    """ 
    Specify what is to be plotted at each frame.
    Input:  plotdata, an instance of clawpack.visclaw.data.ClawPlotData.
    Output: a modified version of plotdata.
    
    """ 

    if plotdata is None:
        from clawpack.visclaw.data import ClawPlotData
        plotdata = ClawPlotData()


    from clawpack.visclaw import colormaps

    plotdata.clearfigures()  # clear any old figures,axes,items data
    

    # Figure for pressure
    # -------------------

    plotfigure = plotdata.new_plotfigure(name='Pressure', figno=0)

    # Set up for axes in this figure:
    plotaxes = plotfigure.new_plotaxes()
    plotaxes.xlimits = 'auto'
    plotaxes.ylimits = 'auto'
    plotaxes.title = 'Pressure'
    plotaxes.scaled = True      # so aspect ratio is 1

    # Set up for item on these axes:
    plotitem = plotaxes.new_plotitem(plot_type='2d_pcolor')
    plotitem.plot_var = 0
    plotitem.pcolor_cmap = colormaps.blue_yellow_red
    plotitem.pcolor_cmin = -2.0
    plotitem.pcolor_cmax = 2.0
    plotitem.add_colorbar = True
    

    # Figure for scatter plot
    # -----------------------

    plotfigure = plotdata.new_plotfigure(name='scatter', figno=3)
    plotfigure.show = (qref_dir is not None)  # don't plot if 1d solution is missing

    # Set up for axes in this figure:
    plotaxes = plotfigure.new_plotaxes()
    plotaxes.xlimits = [0,1.5]
    plotaxes.ylimits = [-2.,4.]
    plotaxes.title = 'Scatter plot'

    # Set up for item on these axes: scatter of 2d data
    plotitem = plotaxes.new_plotitem(plot_type='1d_from_2d_data')
    
    def p_vs_r(current_data):
        # Return radius of each grid cell and p value in the cell
        from pylab import sqrt
        x = current_data.x
        y = current_data.y
        r = sqrt(x**2 + y**2)
        q = current_data.q
        p = q[0,:,:]
        return r,p

    plotitem.map_2d_to_1d = p_vs_r
    plotitem.plot_var = 0
    plotitem.plotstyle = 'o'
    plotitem.color = 'b'
    plotitem.show = (qref_dir is not None)       # show on plot?
    
    # Set up for item on these axes: 1d reference solution
    plotitem = plotaxes.new_plotitem(plot_type='1d_plot')
    plotitem.outdir = qref_dir
    plotitem.plot_var = 0
    plotitem.plotstyle = '-'
    plotitem.color = 'r'
    plotitem.kwargs = {'linewidth': 2}
    plotitem.show = True       # show on plot?
    
    def make_legend(current_data):
        import matplotlib.pyplot as plt
        plt.legend(('2d data', '1d reference solution'))

    plotaxes.afteraxes = make_legend


    # Parameters used only when creating html and/or latex hardcopy
    # e.g., via clawpack.visclaw.frametools.printframes:

    plotdata.printfigs = True                # print figures
    plotdata.print_format = 'png'            # file format
    plotdata.print_framenos = 'all'          # list of frames to print
    plotdata.print_fignos = 'all'            # list of figures to print
    plotdata.html = True                     # create html files of plots?
    plotdata.html_homelink = '../README.html'   # pointer for top of index
    plotdata.html_movie = 'JSAnimation'      # new style, or "4.x" for old style
    plotdata.latex = True                    # create latex file of plots?
    plotdata.latex_figsperline = 2           # layout of plots
    plotdata.latex_framesperline = 1         # layout of plots
    plotdata.latex_makepdf = False           # also run pdflatex?

    return plotdata

    




"""
Module to create topo and qinit data files for this example.
"""

from __future__ import print_function
from clawpack.geoclaw import topotools
from clawpack.geoclaw.data import Rearth  # radius of earth
from interp import pwcubic
from clawpack.clawutil.data import ClawData
from numpy import *

from mapper import latlong, gcdist

probdata = ClawData()
probdata.read('setprob.data', force=True)
theta_island = probdata.theta_island
print("theta_island = ",theta_island)

(xisland,yisland) = latlong(1600.e3, theta_island, 40., Rearth)
print('Island is centered at: (%.3f, %.3f)' % (xisland,yisland))

def maketopo():
    """
    Output topography file for the entire domain
    and near-shore in one location.
    """
    # 6 arcminute resolution:
    nxpoints=401
    nypoints=401
    xlower=-20.e0
    xupper= 20.e0
    ylower= 20.e0
    yupper= 60.e0
    outfile= "ocean.tt3"
    topotools.topo3writer(outfile,topo,xlower,xupper,ylower,yupper,nxpoints,nypoints)

    # round off lat/lon to 2 digits so even multiple of 0.01 degree = 36 arcsec
    xlower= round(xisland - 1., 2)
    xupper= round(xisland + 1., 2)
    ylower= round(yisland - 1., 2)
    yupper= round(yisland + 1., 2)
    # 18 arcsecond resolution:
    nxpoints = int((xupper-xlower) * 3600. / 18.) + 1
    nypoints = int((yupper-ylower) * 3600. / 18.) + 1
    print('island topofile has %i by %i points' % (nxpoints,nypoints))
    outfile= "island.tt3"
    topotools.topo3writer(outfile,topo,xlower,xupper,ylower,yupper,nxpoints,nypoints)

def makeqinit():
    """
    Create qinit data file
    """
    nxpoints=100
    nypoints=100
    xlower=-20.e0
    xupper= 20.e0
    ylower= 20.e0
    yupper= 60.e0
    outfile= "hump.xyz"
    topotools.topo1writer(outfile,qinit,xlower,xupper,ylower,yupper,nxpoints,nypoints)

def shelf1(r):
    """
    Ocean followed by continental slope, continental shelf, and beach.
    The ocean is flat, the slope is cubic, and the shelf and beach are linear.
    """

    rad1 = 1500e3         # beginning of continental slope
    rad2 = rad1 + 60e3   # end of slope, start of shelf
    rad3 = rad2 + 80e3   # end of shelf, start of beach
    rad4 = rad3 + 5e3    # radius of shoreline where z=0
    z1 = -4000.          # depth from r=0  to rad1 (ocean)
    z2 = -100.           # depth at r=rad2 (start of shelf)
    z3 = -100.           # depth at r=rad3 (start of beach)
    z4 = 0.              # depth at shoreline
    xi = array([0., rad1, rad2, rad3, rad4])
    zl = array([z1, z1, z2, z3, z4])
    zr = zl  # continuous!
    slope_of_shelf = (z3 - z2) / (rad3 - rad2)
    slope_of_beach = (z4 - z3) / (rad4 - rad3)
    print("Slope of shelf = ",slope_of_shelf)
    print("Slope of beach = ",slope_of_beach)
    slopel = array([0., 0., slope_of_shelf, slope_of_shelf, slope_of_beach])
    sloper = array([0., 0., slope_of_shelf, slope_of_beach, slope_of_beach])
    z = pwcubic(xi, zl, zr, slopel, sloper, r)
    return z

def island1(r):
    """
    Island created using piecewise cubic with radius rad with zero slope at peak and
    base. Raises topo by ztop at the peak.
    """
    rad = 30e3
    ztop = 120.  # 20m above sealevel on the 100m deep shelf.
    xi = array([0., rad])
    zl = array([ztop, 0.])
    zr = zl
    slopel = array([0., 0.])
    sloper = slopel
    z = pwcubic(xi, zl, zr, slopel, sloper, r)
    return z

def topo(x,y):
    """
    x = longitude, y = latitude in degrees
    """
    import numpy as np
    x0 = 0.0
    y0 = 40.0
    d = gcdist(x0,y0,x,y,Rearth)
    z = shelf1(d)
    d = gcdist(xisland,yisland,x,y,Rearth)
    z = z + island1(d)
    z = where(z>200, 200, z)   # cut off topo at 200 m above sea level
    return z


def qinit(x,y):
    """
    Gaussian hump:
    """
    from numpy import where
    x0 = 0.0
    y0 = 40.0
    d = gcdist(x0,y0,x,y,Rearth)
    ze = -0.5e-9 * d**2
    z = where(ze>-100., 20.e0*exp(ze), 0.)
    return z

if __name__=='__main__':
    maketopo()
    makeqinit()



def test_advection_reaction():
    """ tests against expected sharpclaw results """
    from clawpack.pyclaw.examples.advection_reaction_2d import advection_reaction
    from clawpack.pyclaw.util import test_app, check_diff

    def verify_advection_reaction(controller):
        """ given an expected value, returns a verification function """
        import numpy as np
        import os

        test_solution = controller.solution.state.get_q_global()

        if test_solution is not None:
            thisdir = os.path.dirname(__file__)
            expected_density = np.loadtxt(os.path.join(thisdir,'advection_reaction.txt'))
            test_density = test_solution[0,:,:]
            return check_diff(expected_density, test_density, reltol=1.e-5,delta=controller.solution.grid.delta)

    return test_app(advection_reaction.setup, verify_advection_reaction, {"outdir":''})


# if __name__=="__main__":
#     import nose
#     nose.main()



#!/usr/bin/env python
from __future__ import absolute_import
from __future__ import print_function
import sys
from six.moves import range

try:
    import numpy as np
    from petsc4py import PETSc
except:
    sys.path.append("/opt/share/ksl/petsc4py/dev-aug29/ppc450d/lib/python/")
    sys.path.append("/opt/share/ksl/numpy/dev-aug29/ppc450d/lib/python/")
    
    import numpy as np
    from petsc4py import PETSc

class PetCLAW:
    def advection1D(self, M, cfl, T):
        '''Script to solve 1D advection equation:
        q_t + q_x = 0
        Using first-order finite differences'''
        
        da = PETSc.DA().create([M])
        da.setUniformCoordinates() # solves the problem from 0 to 1
        da.view()

        xvec = da.getCoordinates()
        xvec.view()
        x = xvec.getArray()
        
        h = x[1]-x[0]
        k=cfl*h
        
        fg = da.createGlobalVector()
        fl = da.createLocalVector()
        
        N=int(round(T/k))
        
        # Initial condition:
        q = np.exp(-10*(x-0.5)**2)
        fg.setArray(q)
        da.globalToLocal(fg,fl)

        fg.view()
        
        for n in range(N+1):
            q = fl.getArray()
            q[1:]=q[1:]-cfl*(q[1:]-q[:-1])
            fl.setArray(q)
            da.localToGlobal(fl,fg)            
            fg.view()
            da.globalToLocal(fg,fl)

    def run(self):
        OptDB = PETSc.Options()
        M = OptDB.getInt('M', 16)
        cfl = OptDB.getReal('cfl',0.95)
        T = OptDB.getReal('T',2.)
        self.advection1D(M, cfl, T)
        print('Done')
        return
          
if __name__ == '__main__':
    PetCLAW().run()
          

    
    



#!/usr/bin/env python
# encoding: utf-8

import numpy

from clawpack.geoclaw.units import units, convert


def test_conversions(verbose=False):
    r"""Test unit conversions."""

    for (measurement_type, measurement_units) in units.items():
        value = numpy.pi
        units_list = list(units[measurement_type].keys())
        for i in range(len(units_list)):
            if verbose:
                print("%s (%s) -> (%s)" % (value, units_list[i - 1], 
                                                  units_list[i]))
            value = convert(value, units_list[i - 1], units_list[i])
        numpy.testing.assert_allclose([value], [numpy.pi],
                      err_msg="Measurement tyep %s failed." % measurement_type)

if __name__ == '__main__':
    test_conversions(verbose=True)



"""
    Download topo and dtopo files needed for this example.
    
    Call functions with makeplots==True to create plots of topo, slip, and dtopo.
"""

from __future__ import print_function
import os,sys
import clawpack.clawutil.data
from clawpack.geoclaw import topotools
from numpy import *

try:
    CLAW = os.environ['CLAW']
except:
    raise Exception("*** Must first set CLAW enviornment variable")

# Scratch directory for storing topo and dtopo files:
scratch_dir = os.path.join(CLAW, 'geoclaw', 'scratch')


# Initial data for adjoint is Gaussian hump around this location:
# DART 32412 location:
#xcenter = -86.392
#ycenter = -17.975

# For regression test, move closer:
xcenter = -76.
ycenter = -36.

def get_topo(makeplots=False):
    """
    Retrieve the topo file from the GeoClaw repository.
    """
    from clawpack.geoclaw import topotools
    topo_fname = 'etopo10min120W60W60S0S.asc'
    url = 'http://depts.washington.edu/clawpack/geoclaw/topo/etopo/' + topo_fname
    clawpack.clawutil.data.get_remote_file(url, output_dir=scratch_dir, 
            file_name=topo_fname, verbose=True)

    if makeplots:
        from matplotlib import pyplot as plt
        topo = topotools.Topography(os.path.join(scratch_dir,topo_fname), 
                                    topo_type=2)
        topo.plot()
        fname = os.path.splitext(topo_fname)[0] + '.png'
        plt.savefig(fname)
        print("Created ",fname)


def makeqinit():
    """
        Create qinit data file
    """
    
    nxpoints = 201
    nypoints = 201
    
    xlower = xcenter - 1.5
    xupper = xcenter + 1.5
    ylower = ycenter - 1.5
    yupper = ycenter + 1.5
    
    outfile= "hump.xyz"     
    topotools.topo1writer(outfile,qinit,xlower,xupper,ylower,yupper,nxpoints,nypoints)

def qinit(x,y):
    from numpy import where
    from clawpack.geoclaw.util import haversine

    #radius = 1.0e0
    #ze = sqrt((x-xcenter)**2 + (y-ycenter)**2)
    #z = where(ze<radius, 1.0e0, 0.e0)

    # Gaussian using distance in meters:
    r = haversine(x,y,xcenter,ycenter)
    z = exp(-(r/20e3)**2)
    return z

if __name__=='__main__':
    get_topo(False)
    makeqinit()



#!/usr/bin/env python
"""
Module frametools for plotting frames of time-dependent data.
"""

import os
import sys
import traceback
import glob
import time
import types

# "reload" is only available from a module in Python 3.
if sys.version_info[0] >= 3:
    if sys.version_info[1] >= 4:
        from importlib import reload
    else:
        from imp import reload

import clawpack.clawutil.data as clawdata
from clawpack.visclaw.data import ClawPlotData
from clawpack.visclaw import setplot_default
from clawpack.pyclaw import Solution

import numpy as np
import matplotlib.pyplot as plt
from numpy import ma
from clawpack.visclaw import colormaps
from matplotlib.colors import Normalize, LightSource

# This routine is from matplotlib/lib/matplotlib/colors.py
# from v3.1.1.
# https://matplotlib.org/3.1.1/_modules/matplotlib/colors.html
def _vector_magnitude(arr):
    sum_sq = 0
    for i in range(arr.shape[-1]):
        sum_sq += np.square(arr[..., i, np.newaxis])
    return np.sqrt(sum_sq)


#==============================================================================
def plotframe(frameno, plotdata, verbose=False, simple=False, refresh=False):
#==============================================================================
    """
    Plot a single frame from the computation.

    This routine checks input, loads the appropriate frame from each
    outdir, and then calls plot_frame.

    If simple == True, then plot_frame is not called.  In this case,
    plotdata.setplot should actually be a (matplotlib) function that takes a
    Solution object and plots it.
    """

    if verbose:  print('    Plotting frame %s ... '  % frameno)

    if simple:
        plotfun = plotdata.setplot
        sol = Solution(frameno,path=plotdata.outdir,file_format=plotdata.format)
        plotfun(sol)
        return

    try:
        plotfigure_dict = plotdata.plotfigure_dict
    except AttributeError:
        print('*** Error in plotframe: plotdata missing plotfigure_dict')
        print('*** This should not happen')
        return None

    if len(plotfigure_dict) == 0:
        print('*** Warning in plotframe: plotdata has empty plotfigure_dict')
        print('*** Apparently no figures to plot')

    plotdata.set_outdirs()  # set _outdirs attribute to be list of
                                      # all outdirs for all items

    framesolns = []
    # loop over all outdirs:
    if len(plotdata._outdirs) == 0:
        plotdata._outdirs = [plotdata.outdir]
    for outdir in plotdata._outdirs:
        framesolns.append(plotdata.getframe(frameno, outdir, refresh=refresh))


    # It seems that 'current_data', returned from plot_frame,
    # doesn't get used at all.
    plot_frame(framesolns, plotdata, frameno,verbose=verbose)


#==============================================================================
def plot_frame(framesolns,plotdata,frameno=0,verbose=False):
#==============================================================================
    """
    Plot a single frame from the computation.

    This routine checks input and then calls plotitemN for the
    proper space dimension.  Framesolns is a list of solutions at the same time
    but usually from different output directories.
    """

    if type(framesolns) is not list:
        framesolns = [framesolns]

    t = framesolns[0].t

    # initialize current_data, which will be passed
    # to afterframe, afteraxes, afterpatch commands
    current_data = clawdata.ClawData()
    current_data.add_attribute('user',{})   # for user specified attributes
                                            # to avoid potential conflicts
    current_data.add_attribute('plotdata',plotdata)
    current_data.add_attribute('frameno',frameno)
    current_data.add_attribute('t',t)
    current_data.add_attribute('var',None)
    current_data.add_attribute('plotaxes',None)
    current_data.add_attribute('plotfigure',None)

    # call beforeframe if present, which might define additional
    # attributes in current_data or otherwise set up plotting for this
    # frame.
    beforeframe =  getattr(plotdata, 'beforeframe', None)
    current_data = run_str_or_func(beforeframe,current_data)


    # iterate over each single plot that makes up this frame:
    # -------------------------------------------------------

    if plotdata.mode() == 'iplotclaw':
        plt.ion()
        print('    Plotting Frame %s at t = %s' % (frameno,t))
        requested_fignos = plotdata.iplotclaw_fignos
    else:
        requested_fignos = plotdata.print_fignos

    plotted_fignos = []
    kml_fignos = []

    plotdata = set_show(plotdata)   # set _show attributes for which figures
                                    # and axes should be shown.

    # loop over figures to appear for this frame:
    # -------------------------------------------

    for figname in plotdata._fignames:
        plotfigure = plotdata.plotfigure_dict[figname]
        if (not plotfigure._show) or (plotfigure.type != 'each_frame'):
            continue  # skip to next figure

        figno = plotfigure.figno
        if (requested_fignos != 'all') and (figno not in requested_fignos):
            continue # skip to next figure

        plotted_fignos.append(figno)
        if plotfigure.use_for_kml:
            kml_fignos.append(figno)

        if (plotfigure.facecolor is None) and \
           ('facecolor' not in plotfigure.kwargs):
            # Use white as default starting in v5.10.0
            # To use old default Clawpack tan, in setplot.py set:
            #     plotfigure.facecolor = \
            #           clawpack.visclaw.colormaps.clawpack_tan
            plotfigure.kwargs['facecolor'] = 'w'
        elif plotfigure.facecolor is not None:
            plotfigure.kwargs['facecolor'] = plotfigure.facecolor

        if plotfigure.figsize is not None:
            plotfigure.kwargs['figsize'] = plotfigure.figsize

        # create figure and set handle:
        plotfigure._handle = plt.figure(num=figno, **plotfigure.kwargs)

        #plt.ioff()
        if plotfigure.clf_each_frame:
            plt.clf()

        plotaxes_dict = plotfigure.plotaxes_dict

        if (len(plotaxes_dict) == 0) or (len(plotfigure._axesnames) == 0):
            print('*** Warning in plotframe: plotdata has empty plotaxes_dict')
            print('*** Apparently no axes to plot in figure ',figno)

        # loop over axes to appear on this figure:
        # ----------------------------------------

        for axesname in plotfigure._axesnames:
            plotaxes = plotaxes_dict[axesname]
            if not plotaxes._show:
                continue   # skip this axes if no items show

            # create the axes:
            axescmd = getattr(plotaxes,'axescmd','subplot(1,1,1)')
            axescmd = 'plotaxes._handle = plt.%s' % axescmd
            exec(axescmd)

            current_data.plotaxes = plotaxes
            current_data.plotfigure = plotaxes._plotfigure

            beforeaxes = getattr(plotaxes,'beforeaxes',None)
            current_data = run_str_or_func(beforeaxes,current_data)
                
            skip_patches_outside_xylimits = plotaxes.skip_patches_outside_xylimits
            
            if skip_patches_outside_xylimits is None:
                # User didn't set.  Set to True unless there's a mapped grid
                
                mapc2p_exists = (plotdata.mapc2p is not None)
                if not mapc2p_exists:
                    # check every item in case there's a mapc2p:
                    for itemname in plotaxes._itemnames:
                        plotitem = plotaxes.plotitem_dict[itemname]
                        mapc2p_exists = mapc2p_exists or \
                                           (plotitem.mapc2p is not None)
                                           
                skip_patches_outside_xylimits = not mapc2p_exists
                

            # NOTE: This was rearranged December 2009 to
            # loop over patches first and then over plotitems so that
            # finer patches will plot over coarser patches of other items.
            # Needed in particular if masked arrays are used, e.g. if
            # two different items do pcolor plots with different color maps
            # for different parts of the domain (e.g. water and land).

            for i, framesoln in enumerate(framesolns):

                if abs(framesoln.t - t) > 1e-12:
                    print('*** Warning: t values do not agree for frame ',frameno)
                    print('*** t = %22.15e for outdir = %s' % (t,plotdata.outdir))
                    print('*** t = %22.15e for outdir = %s' % (framesoln.t,plotdata._outdirs[i]))

                current_data.add_attribute('framesoln',framesoln)

                # loop over patches:
                # ----------------


                num_skipped = 0
                
                for stateno,state in enumerate(framesoln.states):

                    patch = state.patch

                    if skip_patches_outside_xylimits:
                        # skip patches not visible based on xlimits,ylimits:
                        if (plotaxes.xlimits is not None) \
                                & (type(plotaxes.xlimits) is not str):
                            if (patch.dimensions[0].lower \
                                        >= plotaxes.xlimits[1]) \
                                    or  (patch.dimensions[0].upper \
                                        <= plotaxes.xlimits[0]):
                                num_skipped += 1
                                continue  # go to next patch
    
                        if len(patch.dimensions) > 1:
                            # 2d patch
                            if (plotaxes.ylimits is not None) \
                                    & (type(plotaxes.ylimits) is not str):
                                if (patch.dimensions[1].lower \
                                            >= plotaxes.ylimits[1]) \
                                        or  (patch.dimensions[1].upper \
                                            <= plotaxes.ylimits[0]):
                                    num_skipped += 1
                                    continue  # go to next patch

                    current_data.add_attribute('patch',patch)
                    current_data.add_attribute("level",1)
                    current_data.add_attribute('q',state.q)
                    current_data.add_attribute('aux',state.aux)
                    current_data.add_attribute('xlower',patch.dimensions[0].lower)
                    current_data.add_attribute('xupper',patch.dimensions[0].upper)

                    current_data.add_attribute("x",patch.grid.p_centers[0])
                    current_data.add_attribute("dx",patch.delta[0])


                    if patch.num_dim == 2:
                        current_data.add_attribute('ylower',patch.dimensions[1].lower)
                        current_data.add_attribute('yupper',patch.dimensions[1].upper)
                        current_data.add_attribute('y',patch.grid.p_centers[1])
                        current_data.add_attribute('dy',patch.delta[1])

                    if plotfigure.use_for_kml:

                        # -------------------------------------------------------------
                        # Mask out all coarse grid regions that are under fine grids
                        # This could be made into a subroutine
                        # -------------------------------------------------------------
                        this_level = patch.level

                        xc_centers,yc_centers = patch.grid.c_centers
                        mask_coarse = np.empty(xc_centers.shape, dtype=bool)
                        mask_coarse.fill(False)

                        # iterate over all grids to see which one masks this grid
                        for stateno_fine,state_fine in enumerate(framesoln.states):
                            # iterate over all patches, and find any finer level grids that are
                            # sitting on top of this patch/grid/state.
                            patch_fine = state_fine.patch

                            # Only look at patches one level finer
                            if patch_fine.level != this_level+1:
                                continue

                            xlower_fine = patch_fine.dimensions[0].lower
                            xupper_fine = patch_fine.dimensions[0].upper
                            ylower_fine = patch_fine.dimensions[1].lower
                            yupper_fine = patch_fine.dimensions[1].upper

                            m1 = (xc_centers > xlower_fine) & (xc_centers < xupper_fine)
                            m2 = (yc_centers > ylower_fine) & (yc_centers < yupper_fine)

                            # Mask all fine grid regions
                            mask_coarse = (m1 & m2) | mask_coarse

                        current_data.add_attribute('mask_coarse',mask_coarse)
                        # -------------------------------------------------------------
                        # Done with masking
                        # -------------------------------------------------------------


                    # loop over items:
                    # ----------------

                    for itemname in plotaxes._itemnames:

                        plotitem = plotaxes.plotitem_dict[itemname]

                        try:
                            item_outdir = plotitem.outdir
                            if not plotitem.outdir:
                                item_outdir = plotdata.outdir
                            if item_outdir != plotdata._outdirs[i]:
                                # skip to next item
                                continue
                        except (NameError, AttributeError):  # no outdir if plotting from memory
                            pass

                        num_dim = plotitem.num_dim

                        # option to suppress printing some levels:
                        amr_data_show = plotitem.amr_data_show
                        if len(amr_data_show) > 0:
                            j = min(len(amr_data_show), patch.level) - 1
                            show_this_level = amr_data_show[j]
                        else:
                            show_this_level = True

                        if plotitem._show and show_this_level:
                            if num_dim == 1:
                                plotitem_fun = plotitem1
                            elif num_dim == 2:
                                plotitem_fun = plotitem2
                            current_data = plotitem_fun(framesoln,plotitem,current_data,stateno)

                            if verbose:
                                print('      Plotted  plotitem ', itemname)

                    # end of loop over plotitems
                # end of loop over patches

            if False and num_skipped > 0:
                # possible warning message:
                print('Skipped plotting %i patches not visible' % num_skipped)

            # end of loop over framesolns


            for itemname in plotaxes._itemnames:
                plotitem = plotaxes.plotitem_dict[itemname]
                if plotitem.afteritem:
                    print("*** ClawPlotItem.afteritem is deprecated")
                    print("*** use ClawPlotAxes.afteraxes ")
                    print("*** or  ClawPlotItem.afterpatch instead")
                try:
                    if plotfigure.use_for_kml:
                        # Build colorbar for Google Earth (now done in setplot.py)
                        pass
                    elif plotitem.has_attribute('add_colorbar') and plotitem.add_colorbar:
                        pobj = plotitem._current_pobj # most recent plot object
                        # set dictionary values for keywords explicitly set:
                        if plotitem.colorbar_shrink is not None:
                            plotitem.colorbar_kwargs['shrink'] = \
                                    plotitem.colorbar_shrink
                        if plotitem.colorbar_ticks is not None:
                            plotitem.colorbar_kwargs['ticks'] = \
                                    plotitem.colorbar_ticks
                        if plotitem.colorbar_extend is not None:
                            plotitem.colorbar_kwargs['extend'] = \
                                    plotitem.colorbar_extend

                        cbar = plt.colorbar(pobj, **plotitem.colorbar_kwargs)

                        if plotitem.has_attribute('colorbar_tick_labels'):
                            if plotitem.colorbar_tick_labels is not None:
                                cbar.ax.set_yticklabels(plotitem.colorbar_tick_labels)
                        if plotitem.colorbar_label is not None:
                            cbar.set_label(plotitem.colorbar_label)
                except:
                    print("*** problem generating colorbar")
                    pass

            if plotfigure.use_for_kml:
                pass
            else:
                if plotaxes.title_with_t:
                    if 'd:h:m:s' in plotaxes.title:
                        #from datetime import timedelta
                        #t_str = str(timedelta(seconds=t))
                        #title_str = plotaxes.title.replace('d:h:m:s',t_str)

                        # formats the same as above but doesn't use datetime:
                        days, remainder = divmod(t, 24*3600)
                        hours, remainder = divmod(remainder, 3600)
                        minutes, seconds = divmod(remainder, 60)
                        t_str = '%i days, %i:%s:%s' \
                                           % (days,hours,\
                                              str(int(minutes)).zfill(2),\
                                              str(int(seconds)).zfill(2))
                        title_str = plotaxes.title.replace('d:h:m:s',t_str)

                    elif 'h:m:s' in plotaxes.title:
                        # keep total hours, not days
                        hours, remainder = divmod(t, 3600)
                        minutes, seconds = divmod(remainder, 60)
                        t_str = '%i:%s:%s' % (hours,\
                                              str(int(minutes)).zfill(2),\
                                              str(int(seconds)).zfill(2))
                        title_str = plotaxes.title.replace('h:m:s',t_str)

                    elif plotaxes.title_t_format:
                        # now allow user to specify other formats for t:
                        t_str = plotaxes.title_t_format % t

                        title_str = "%s at time t = %s" \
                                  % (plotaxes.title,t_str)
                        
                    elif (t==0.) | ((t>=0.001) & (t<1000.)):
                        title_str = "%s at time t = %14.8f" \
                                  % (plotaxes.title,t)
                    else:
                        title_str = "%s at time t = %14.8e" \
                                  % (plotaxes.title,t)


                else:
                    # omit t from title:
                    title_str = plotaxes.title

                if plotaxes.title_fontsize is not None:
                    plotaxes.title_kwargs['fontsize'] = plotaxes.title_fontsize
                plt.title(title_str, **plotaxes.title_kwargs)

            # call an afteraxes function if present:
            afteraxes =  getattr(plotaxes, 'afteraxes', None)
            current_data = run_str_or_func(afteraxes,current_data)

            if plotaxes.scaled:
                plt.axis('scaled')
            elif plotaxes.image:
                plt.axis('image')

            # set axes limits:
            if (plotfigure.use_for_kml and plotfigure.kml_use_figure_limits):
                if (plotfigure.kml_xlimits is not None) & (type(plotfigure.kml_xlimits) is not str):
                    try:
                        plt.xlim(plotfigure.kml_xlimits[0], plotfigure.kml_xlimits[1])
                    except:
                        print(" ")
                        print("*** KML error : Set plotfigure.kml_xlimits")
                        print(" ")
                        return

                if (plotfigure.kml_ylimits is not None) & (type(plotfigure.kml_ylimits) is not str):
                    try:
                        plt.ylim(plotfigure.kml_ylimits[0], plotfigure.kml_ylimits[1])
                    except:
                        print(" ")
                        print("*** KML error : Set plotfigure.kml_ylimits")
                        print(" ")
                        return
            else:
                if (plotaxes.xlimits is not None) & (type(plotaxes.xlimits) is not str):
                    try:
                        plt.xlim(plotaxes.xlimits[0], plotaxes.xlimits[1])
                    except:
                        pass  # let axis be set automatically
                if (plotaxes.ylimits is not None) & (type(plotaxes.ylimits) is not str):
                    try:
                        plt.ylim(plotaxes.ylimits[0], plotaxes.ylimits[1])
                    except:
                        pass  # let axis be set automatically

            if plotaxes.useOffset is not None:
                plt.ticklabel_format(useOffset = plotaxes.useOffset)

            if plotaxes.grid:
                plt.grid(**plotaxes.grid_kwargs)

            if plotaxes.xticks_fontsize is not None:
                plotaxes.xticks_kwargs['fontsize'] = plotaxes.xticks_fontsize
            if plotaxes.xticks_kwargs != {}:
                plt.xticks(**plotaxes.xticks_kwargs)

            if plotaxes.yticks_fontsize is not None:
                plotaxes.yticks_kwargs['fontsize'] = plotaxes.yticks_fontsize
            if plotaxes.yticks_kwargs != {}:
                plt.yticks(**plotaxes.yticks_kwargs)

            if plotaxes.xlabel is not None:
                if plotaxes.xlabel_fontsize is not None:
                    plotaxes.xlabel_kwargs['fontsize'] = plotaxes.xlabel_fontsize
                plt.xlabel(plotaxes.xlabel, **plotaxes.xlabel_kwargs)
            if plotaxes.ylabel is not None:
                if plotaxes.ylabel_fontsize is not None:
                    plotaxes.ylabel_kwargs['fontsize'] = plotaxes.ylabel_fontsize
                plt.ylabel(plotaxes.ylabel, **plotaxes.ylabel_kwargs)

            if plotaxes.aspect_latitude is not None:
                plt.gca().set_aspect(1./np.cos(plotaxes.aspect_latitude \
                            * np.pi/180))
            elif plotaxes.aspect is not None:
                plt.gca().set_aspect(plotaxes.aspect)

            # end of loop over plotaxes
        # end of loop over plotfigures


    # call an afterframe function if present:
    afterframe =  getattr(plotdata, 'afterframe', None)
    current_data = run_str_or_func(afterframe,current_data)

    if plotdata.mode() == 'iplotclaw':
        plt.ion()
    for figno in plotted_fignos:
        plt.figure(figno)
        plt.draw()

    if verbose:
        print('    Done with plotframe for frame %i at time %g' % (frameno,t))


    # print the figure(s) to file(s) if requested:
    if (plotdata.mode() != 'iplotclaw') & plotdata.printfigs:
        # iterate over all figures that are to be printed:
        for figname in plotdata._fignames:
            plotfigure = plotdata.plotfigure_dict[figname]
            figno = plotfigure.figno
            if figno in plotted_fignos:
                if figno in kml_fignos:
                    printfig(frameno=frameno, figno=figno,\
                             file_prefix=plotdata.file_prefix,\
                             format=plotdata.print_format, plotdir=plotdata.plotdir,\
                             verbose=verbose,kml_fig=True,kml_dpi=plotfigure.kml_dpi,
                             kml_figsize=plotfigure.kml_figsize)
                else:
                    printfig(frameno=frameno, figno=figno,\
                             file_prefix=plotdata.file_prefix,\
                             format=plotdata.print_format, plotdir=plotdata.plotdir,\
                             verbose=verbose,kml_fig=False)

    return current_data
    # end of plotframe


def run_str_or_func(str_or_func,current_data):
    if str_or_func is None:
        return current_data
    if isinstance(str_or_func, str):
        exec(str_or_func)
    else:
        output = str_or_func(current_data)
        if output:
            return output
    return current_data


def params_dict(plotitem, base_params, level_params, level):
    """
    Create a dictionary containing the plot parameters.

    For each level_param check if there is an amr_ list for this
    parameter and if so select the value corresponding to the level of
    this patch.  Otherwise, use plotitem attribute of this name.
    """
    pp = {}
    for plot_param in base_params:
        pp[plot_param] = getattr(plotitem, plot_param, None)


    for plot_param in level_params:
        amr_plot_param = "amr_%s" % plot_param
        amr_list = getattr(plotitem, amr_plot_param, [])
        if len(amr_list) > 0:
            index = min(level, len(amr_list)) - 1
            pp[plot_param] = amr_list[index]
        else:
            pp[plot_param] = getattr(plotitem, plot_param, None)

    return pp

#==================================================================
def plotitem1(framesoln, plotitem, current_data, stateno):
#==================================================================
    """
    Make a 1d plot for a single plot item for the solution in framesoln.

    The current_data object holds data that should be passed into
    afterpatch or afteraxes if these functions are defined.  The functions
    may add to this object, so this function should return the possibly
    modified current_data for use in other plotitems or in afteraxes or
    afterframe.

    """

    plotdata = plotitem._plotdata

    state = framesoln.states[stateno]
    patch = state.patch
    current_data.patch = patch
    current_data.q = state.q
    current_data.aux = state.aux

    # level not currently set in 1d, default to level = 1:
    current_data.level = getattr(patch,'level',1)

    t = framesoln.t

    # the following plot parameters should be set and independent of
    # which AMR level a patch is on:

    base_params = ['plot_type','afteritem','mapc2p','MappedGrid','gaugeno']

    # the following plot parameters may be set, depending on what
    # plot_type was requested:

    # Note: not all of these are initialized in data.py - need to clean up
    level_params = ['plot_var','afterpatch','plotstyle','color','kwargs',\
             'plot_var2','fill_where','map_2d_to_1d','data_show']

    pp = params_dict(plotitem, base_params, level_params, patch.level)

    if pp['mapc2p'] is None:
        # if this item does not have a mapping, check for a global mapping:
        pp['mapc2p'] = getattr(plotdata, 'mapc2p', None)

    if pp['plot_type'] == '1d':
        pp['plot_type'] = '1d_plot'  # '1d' is deprecated

    if pp['plot_type'] == '1d_gauge_trace':
        gaugesoln = plotdata.getgauge(pp['gaugeno'])
        xc_centers = None
        xc_edges = None

    elif pp['plot_type'] in ['1d_plot', '1d_semilogy', '1d_fill_between',
            '1d_empty','1d_from_2d_data', '1d_pwconst']:
        var = get_var(state,pp['plot_var'],current_data)
        current_data.var = var

    else:
        raise ValueError("Unrecognized plot_type: %s" % pp['plot_type'])

    if pp['plot_type'] == '1d_fill_between':
        var  = get_var(state,pp['plot_var'],current_data)
        var2 = get_var(state,pp['plot_var2'],current_data)
        current_data.var = var
        current_data.add_attribute('var2',var2)

    # Grid mapping:
    xc_centers = patch.grid.c_centers

    if pp['MappedGrid'] is None:
        pp['MappedGrid'] = (pp['mapc2p'] is not None)

    if (pp['MappedGrid'] & (pp['mapc2p'] is None)):
        print("*** Warning: MappedGrid == True but no mapc2p specified")
    elif pp['MappedGrid']:
        p_centers = pp['mapc2p'](xc_centers[0])
    else:
        p_centers = xc_centers[0]

    if pp['plot_type'] == '1d_from_2d_data':
        if not pp['map_2d_to_1d']:
            print('*** Error, plot_type = 1d_from_2d_data requires ')
            print('*** map_2d_to_1d function as plotitem attribute')
            raise
            return
        p_centers, var = pp['map_2d_to_1d'](current_data)
        p_centers = p_centers.flatten()  # convert to 1d
        var = var.flatten()  # convert to 1d

    # The plot commands using matplotlib:

    if pp['color']:
        pp['kwargs']['color'] = pp['color']

    if pp['data_show']:
        if (pp['plot_type'] in ['1d_plot','1d_from_2d_data']):
            pobj=plt.plot(p_centers,var,pp['plotstyle'],**pp['kwargs'])

        elif pp['plot_type'] == '1d_semilogy':
            pobj=plt.semilogy(p_centers,var,pp['plotstyle'], **pp['kwargs'])

        elif pp['plot_type'] == '1d_fill_between':
            if pp['fill_where']:
                pp['fill_where'] = pp['fill_where'].replace('plot_var','var')
                pp['fill_where'] = pp['fill_where'].replace('plot_var2','var2')
                pobj=plt.fill_between(p_centers,var,var2,pp['fill_where'],**pp['kwargs'])
            else:
                pobj=plt.fill_between(p_centers,var,var2,**pp['kwargs'])

        elif pp['plot_type'] == '1d_pwconst':

            # should improve this to also work on mapped grid
            xc_edges = patch.grid.c_nodes[0]
            medges = len(xc_edges)
            mcells = medges - 1  # number of grid cells
            # double up points for pw constant plot
            xc_edges2 = np.reshape(np.vstack((xc_edges,xc_edges)).T, 
                (2*medges,))
            xc_edges2 = xc_edges2[1:-1]  # discard first and last
            var2 = np.reshape(np.vstack((var,var)).T, (2*mcells,))
            pobj=plt.plot(xc_edges2,var2,pp['plotstyle'],**pp['kwargs'])

        elif pp['plot_type'] == '1d_gauge_trace':

            gauget = gaugesoln.t
            gaugeq = gaugesoln.q[3,:]
            pobj=plt.plot(gauget, gaugeq)

            # interpolate to the current time t:
            try:
                i1 = np.where(gauget < t)[0][-1]
                i1 = min(i1,len(gauget)-2)
                slope = (gaugeq[i1+1]-gaugeq[i1]) / (gauget[i1+1]-gauget[i1])
                qt = gaugeq[i1] + slope * (t-gauget[i1])
            except IndexError:
                qt = gaugeq[0]
                print("Warning: t out of range")
            plt.plot([t], [qt], 'ro')

        elif pp['plot_type'] == '1d_empty':
            # no plot to create (user might make one in afteritem or
            # afteraxes)
            pass

        else:
            raise ValueError("Unrecognized plot_type: %s" % pp['plot_type'])
            return None

    # call an afterpatch function if present:
    if pp['afterpatch']:
        if isinstance(pp['afterpatch'], str):
            # a string to be executed
            exec(pp['afterpatch'])
        else:
            # assume it's a function
            try:
                # set values that may be needed in afterpatch:
                #current_data.patchno = patch.patch_index # can get from patch
                current_data.add_attribute('plotitem',plotitem)
                current_data.patch = patch
                current_data.var = var
                current_data.xlower = patch.dimensions[0].lower
                current_data.xupper = patch.dimensions[0].upper
                current_data.x = p_centers # cell centers
                current_data.dx = patch.delta[0]
                output = pp['afterpatch'](current_data)
                if output: current_data = output
            except:
                print('*** Error in afterpatch ***')
                raise

    try:
        plotitem._current_pobj = pobj
    except NameError:
        pass # if no plot was done

    return current_data


#==================================================================
def plotitem2(framesoln, plotitem, current_data, stateno):
#==================================================================
    """
    Make a 2d plot for a single plot item for the solution in framesoln.

    The current_data object holds data that should be passed into
    afterpatch or afteraxes if these functions are defined.  The functions
    may add to this object, so this function should return the possibly
    modified current_data for use in other plotitems or in afteraxes or
    afterframe.

    """

    plotdata = plotitem._plotdata

    state = framesoln.states[stateno]
    patch = state.patch
    level = patch.level #2d only

    current_data.patch = patch
    current_data.q = state.q
    current_data.aux = state.aux
    current_data.add_attribute('level',level) #2d only

    t = framesoln.t

    # The following plot parameters should be set and independent of
    # which AMR level a patch is on:

    base_params = ['plot_type','afteritem','mapc2p','MappedGrid']

    level_params = ['plot_var','afterpatch','kwargs',
             'celledges_show','celledges_color','celledges_linewidth', 
             'patch_bgcolor',
             'patchedges_show','patchedges_color', 'patchedges_linewidth', 
             'add_colorbar', 'pcolor_cmap','pcolor_cmin','pcolor_cmax',
             'imshow_cmap','imshow_cmin','imshow_cmax',
             'imshow_norm', 'imshow_alpha',
             'contour_levels','contour_nlevels','contour_min','contour_max',
             'contour_colors','contour_cmap','contour_show',
             'fill_cmap','fill_cmin','fill_cmax','fill_colors',
             'schlieren_cmap','schlieren_cmin', 'schlieren_cmax',
             'quiver_coarsening','quiver_var_x','quiver_var_y','quiver_key_show',
             'quiver_key_scale','quiver_key_label_x','quiver_key_label_y',
             'quiver_key_scale','quiver_key_units','quiver_key_kwargs',
             'hillshade_cmap','hillshade_vertical_exaggeration',
             'hillshade_azimuth_degree','hillshade_altitude_degree',
             'hillshade_latlon']

    pp = params_dict(plotitem, base_params, level_params, patch.level)

    if pp['mapc2p'] is None:
        # if this item does not have a mapping, check for a global mapping:
        pp['mapc2p'] = getattr(plotdata, 'mapc2p', None)

    # turn patch background color into a colormap for use with pcolor cmd:
    pp['patch_bgcolormap'] = colormaps.make_colormap({0.: pp['patch_bgcolor'], \
                                             1.: pp['patch_bgcolor']})

    var  = get_var(state,pp['plot_var'],current_data)
    current_data.var = var

    # Grid mapping:

    xc_edges, yc_edges = patch.grid.c_nodes
    xc_centers, yc_centers = patch.grid.c_centers
    if pp['MappedGrid'] is None:
        pp['MappedGrid'] = (pp['mapc2p'] is not None)

    if (pp['MappedGrid'] & (pp['mapc2p'] is None)):
        raise Exception("MappedGrid == True but no mapc2p specified")
    elif pp['MappedGrid']:
        X_center, Y_center = pp['mapc2p'](xc_centers, yc_centers)
        X_edge, Y_edge = pp['mapc2p'](xc_edges, yc_edges)
    else:
        X_center, Y_center = xc_centers, yc_centers
        X_edge, Y_edge = xc_edges, yc_edges

    if ma.isMaskedArray(var):
        # If var is a masked array: plotting should work ok unless all
        # values are masked, in which case pcolor complains and there's
        # no need to try to plot.  Check for this case...
        var_all_masked = (ma.count(var) == 0)
    else:
        # not a masked array, so certainly not all masked:
        var_all_masked = False

    # pcolormesh is much faster but cannot be used with masked coordinate arrays
    if ma.isMaskedArray(X_edge) or ma.isMaskedArray(Y_edge):
        pc_cmd = 'pcolor'
        pc_mth = plt.pcolor
    else:
        pc_cmd = 'pcolormesh'
        pc_mth = plt.pcolormesh

    if pp['plot_type'] == '2d_pcolor':

        pcolor_cmd = "plotitem._current_pobj = plt."+pc_cmd+"(X_edge, Y_edge, var, \
                        cmap=pp['pcolor_cmap']"

        if pp['celledges_show']:
            pcolor_cmd += ", edgecolors=pp['celledges_color']"
            pcolor_cmd += ", linewidths=pp['celledges_linewidth']"
        else:
            pcolor_cmd += ", shading='flat'"
        
        if 'rasterized' not in pp['kwargs']:
            pcolor_cmd += ", rasterized=True"
        
        pcolor_cmd += ", **pp['kwargs'])"

        if not var_all_masked:
            exec(pcolor_cmd)


            if (pp['pcolor_cmin'] not in ['auto',None]) and \
                     (pp['pcolor_cmax'] not in ['auto',None]):
                plt.clim(pp['pcolor_cmin'], pp['pcolor_cmax'])
        else:
            #print '*** Not doing pcolor on totally masked array'
            pass

    elif pp['plot_type'] == '2d_imshow':

        if not var_all_masked:
            if pp['imshow_cmin'] in ['auto',None]:
                pp['imshow_cmin'] = np.min(var)
            if pp['imshow_cmax'] in ['auto',None]:
                pp['imshow_cmax'] = np.max(var)

            if pp['imshow_norm'] in ["auto", None]:
                color_norm = Normalize(pp['imshow_cmin'],pp['imshow_cmax'],clip=True)
            else:
                color_norm = pp['imshow_norm']
         
            xylimits = (X_edge[0,0],X_edge[-1,-1],Y_edge[0,0],Y_edge[-1,-1])
            pobj = plt.imshow(np.flipud(var.T), extent=xylimits, \
                    cmap=pp['imshow_cmap'], interpolation='nearest', \
                    norm=color_norm, \
                    alpha=pp["imshow_alpha"]
                    )

            if pp['celledges_show']:
                # This draws cell edges for this level.
                # Note that these will still be visible when imshow is used
                # for finer levels, so imshow doesn't look as good as pcolor
                # when you only want to show celledges on coarse levels.
                # There doesn't seem to be an easy way to fix this.
                pobj = plt.plot(X_edge, Y_edge, color=pp['celledges_color'], 
                                           linewidth=pp['celledges_linewidth'])
                pobj = plt.plot(X_edge.T, Y_edge.T, color=pp['celledges_color'], 
                                           linewidth=pp['celledges_linewidth'])

        else:
            #print '*** Not doing imshow on totally masked array'
            pass


    elif pp['plot_type'] in ['2d_contour','2d_contourf']:
        levels_set = True
        if pp['contour_levels'] is None:
            levels_set = False
            if pp['contour_nlevels'] is None:
                print('*** Error in plotitem2:')
                print('    contour_levels or contour_nlevels must be set')
                raise
                return
            if (pp['contour_min'] is not None) and \
                    (pp['contour_max'] is not None):

                pp['contour_levels'] = np.linspace(pp['contour_min'], \
                       pp['contour_max'], pp['contour_nlevels'])
                levels_set = True


        if pp['celledges_show']:
            pobj = pc_mth(X_edge, Y_edge, np.zeros(var.shape), \
                    cmap=pp['patch_bgcolormap'], 
                    edgecolors=pp['celledges_color'],
                    linewidths=pp['celledges_linewidth'])
        elif pp['patch_bgcolor'] != 'w':
            pobj = pc_mth(X_edge, Y_edge, np.zeros(var.shape), \
                    cmap=pp['patch_bgcolormap'], edgecolors='None')

        if pp['plot_type'] == '2d_contour':
            # create the contour command:
            contourcmd = "pobj = plt.contour(X_center, Y_center, var, "
            if levels_set:
                contourcmd += "pp['contour_levels']"
            else:
                contourcmd += "pp['contour_nlevels']"

            if pp['contour_cmap']:
                if (pp['kwargs'] is None) or ('cmap' not in pp['kwargs']):
                    contourcmd += ", cmap = pp['contour_cmap']"
            elif pp['contour_colors']:
                if (pp['kwargs'] is None) or ('colors' not in pp['kwargs']):
                    contourcmd += ", colors = pp['contour_colors']"

            contourcmd += ", **pp['kwargs'])"

            if (pp['contour_show'] and not var_all_masked):
                # may suppress plotting at coarse levels
                exec(contourcmd)


        if pp['plot_type'] == '2d_contourf':

            # create the contourf command:
            contourcmd = "pobj = plt.contourf(X_center, Y_center, var, "
            if levels_set:
                contourcmd += "pp['contour_levels']"
            else:
                contourcmd += "pp['contour_nlevels']"

            if pp['fill_cmap']:
                if (pp['kwargs'] is None) or ('cmap' not in pp['kwargs']):
                    contourcmd += ", cmap = pp['fill_cmap']"
            elif pp['fill_colors']:
                if (pp['kwargs'] is None) or ('colors' not in pp['kwargs']):
                    contourcmd += ", colors = pp['fill_colors']"


            if (pp['kwargs'] is None) or ('extend' not in pp['kwargs']):
                contourcmd += ", extend = 'both'"

            contourcmd += ", **pp['kwargs'])"


            if (not var_all_masked):
                # may suppress plotting at coarse levels
                exec(contourcmd)

                if pp['fill_cmap'] and \
                         (pp['fill_cmin'] not in ['auto',None]) and \
                         (pp['fill_cmax'] not in ['auto',None]):
                    plt.clim(pp['fill_cmin'], pp['fill_cmax'])


    elif pp['plot_type'] == '2d_patch':
        # plot only the patches, no data:
        if pp['celledges_show']:
            pobj = pc_mth(X_edge, Y_edge, np.zeros(var.shape), \
                    cmap=pp['patch_bgcolormap'], 
                    edgecolors=pp['celledges_color'],
                    linewidths=pp['celledges_linewidth'])
        else:
            pobj = pc_mth(X_edge, Y_edge, np.zeros(var.shape), \
                    cmap=pp['patch_bgcolormap'], shading='flat')


    elif pp['plot_type'] == '2d_schlieren':
        # plot 2-norm of gradient of variable var:

        # No idea why this next line is needed...maybe a 64-/32-bit incompatibility issue?
        var = np.array(var)
        (vx,vy) = np.gradient(var)
        vs = np.sqrt(vx**2 + vy**2)

        pcolor_cmd = "pobj = plt.pcolormesh(X_edge, Y_edge, vs, \
                        cmap=pp['schlieren_cmap']"

        if pp['celledges_show']:
            pcolor_cmd += ", edgecolors=pp['celledges_color']"
        else:
            pcolor_cmd += ", edgecolors='None'"

        pcolor_cmd += ", **pp['kwargs'])"

        if not var_all_masked:
            exec(pcolor_cmd)

            if (pp['schlieren_cmin'] not in ['auto',None]) and \
                     (pp['schlieren_cmax'] not in ['auto',None]):
                plt.clim(pp['schlieren_cmin'], pp['schlieren_cmax'])

    elif pp['plot_type'] == '2d_quiver':
        if pp['quiver_coarsening'] > 0:
            var_x = get_var(state,pp['quiver_var_x'],current_data)
            var_y = get_var(state,pp['quiver_var_y'],current_data)
            Q = plt.quiver(X_center[::pp['quiver_coarsening'],::pp['quiver_coarsening']],
                             Y_center[::pp['quiver_coarsening'],::pp['quiver_coarsening']],
                             var_x[::pp['quiver_coarsening'],::pp['quiver_coarsening']],
                             var_y[::pp['quiver_coarsening'],::pp['quiver_coarsening']],
                             **pp['kwargs'])
                             # units=pp['quiver_units'],scale=pp['quiver_scale'])

            # Show key
            if pp['quiver_key_show']:
                if pp['quiver_key_scale'] is None:
                    key_scale = np.max(np.sqrt(var_x**2+var_y**2))*0.5
                else:
                    key_scale = pp['quiver_key_scale']
                label = r"%s %s" % (str(np.ceil(key_scale)),pp['quiver_key_units'])
                plt.quiverkey(Q,pp['quiver_key_label_x'],pp['quiver_key_label_y'],
                                key_scale,label,**pp['quiver_key_kwargs'])

    elif pp['plot_type'] == '2d_empty':
        # no plot to create (user might make one in afteritem or
        # afteraxes)
        pass

    elif pp['plot_type'] == '2d_hillshade':
        if not var_all_masked:
            if pp['hillshade_latlon']:
                xcoord2meters = 1/111200

                # could eventually adjust dx
                # based on np.cos(np.radians(latitude))

            else:
                xcoord2meters = 1
            
            ve = pp['hillshade_vertical_exaggeration'] * xcoord2meters
            azdeg=pp['hillshade_azimuth_degree']
            altdeg = pp['hillshade_altitude_degree']

            ls = LightSource(azdeg=azdeg, altdeg=altdeg)
            # typically would do the following:
            #  hs = ls.hillshade(z, vert_exag=ve, dx=dx, dy=dy)

            # but here we need to calculate the hillshade by hand so that the values
            # are consistent across all patches. The matplotlib code applies a 
            # contrast stretch that we don't want. 
   
            # This code was modified from:
            # https://matplotlib.org/3.1.1/_modules/matplotlib/colors.html#LightSource
            z = np.flipud(var.T)
            dx = current_data.dx
            dy = current_data.dy
            
            e_dy, e_dx = np.gradient(ve * z, -dy, dx)
            normal = np.empty(z.shape + (3,)).view(type(z))
            normal[..., 0] = -e_dx
            normal[..., 1] = -e_dy
            normal[..., 2] = 1
            normal /= _vector_magnitude(normal)
            intensity = normal.dot(ls.direction)

            # contrast stretch was applied here, code was deleted from the matplotlib routines.

            intensity = np.clip(intensity, 0, 1)
            hs = intensity

            xylimits = (X_edge[0, 0], X_edge[-1, -1], Y_edge[0, 0], Y_edge[-1, -1])
            pobj = plt.imshow(hs, cmap="gray", vmin=0, vmax=1, extent=xylimits)
            color_norm = Normalize(pp['imshow_cmin'],pp['imshow_cmax'],clip=True)

        else:
            #print '*** Not doing hillshade on totally masked array'
            pass

    else:
        raise ValueError("Unrecognized plot_type: %s" % pp['plot_type'])
        return None

    # end of various plot types



    # plot patch patch edges if desired:

    if pp['patchedges_show']:
        for i in [0, X_edge.shape[0]-1]:
            X1 = X_edge[i,:]
            Y1 = Y_edge[i,:]
            plt.plot(X1, Y1, color=pp['patchedges_color'], linewidth=pp['patchedges_linewidth'])
        for i in [0, X_edge.shape[1]-1]:
            X1 = X_edge[:,i]
            Y1 = Y_edge[:,i]
            plt.plot(X1, Y1, color=pp['patchedges_color'], linewidth=pp['patchedges_linewidth'])


    if pp['afterpatch']:
        try:
            if isinstance(pp['afterpatch'], str):
                exec(pp['afterpatch'])
            else:
                # assume it's a function
                #current_data.patchno = patch.patch_index # can get from patch
                current_data.add_attribute('plotitem',plotitem)
                current_data.patch = patch
                current_data.var = var
                current_data.xlower = patch.dimensions[0].lower
                current_data.xupper = patch.dimensions[0].upper
                current_data.ylower = patch.dimensions[1].lower
                current_data.yupper = patch.dimensions[1].upper
                current_data.x = X_center # cell centers
                current_data.y = Y_center # cell centers
                current_data.dx = patch.delta[0]
                current_data.dy = patch.delta[1]

                output = pp['afterpatch'](current_data)
                if output: current_data = output
        except:
            print('*** Warning: could not execute afterpatch')
            raise


    try:
        plotitem._current_pobj = pobj
    except NameError:
        pass # if no pobj generated


    return current_data

#--------------------------------------
def get_var(state, plot_var, current_data):
#--------------------------------------
    """
    Return arrays for spatial variable(s) (on mapped patch if necessary)
    and variable to be plotted on a single patch in num_dim space dimensions.
    """

    if isinstance(plot_var, int):
        var = state.q[plot_var,...]
    else:
        try:
            var = plot_var(current_data)
        except:
            print('*** Error applying function plot_var = ',plot_var)
            raise

    return var


#------------------------------------------------------------------------
def printfig(fname='',frameno='', figno='', file_prefix='fort',
             format='png', plotdir='.',
             verbose=True, kml_fig=False, kml_dpi=None, kml_figsize=None,
             bbox_inches='tight',close_fig=True):
#------------------------------------------------------------------------
    """
    Save the current plot to file fname or standard name from frame/fig.
.
    If fname is nonempty it is used as the filename, with extension
    determined by format if it does not already have a valid extension.

    If fname=='' then save to file frame000NfigJ.ext  where N is the frame
    number frameno passed in, J is the figure number figno passed in,
    and the extension ext is determined by format.
    If figno='' then the figJ part is omitted.
    """

    if file_prefix == 'fort':
        # usual case:
        png_prefix = 'frame'
    else:
        # e.g. file_prefix == 'fgout0001'
        png_prefix = file_prefix + 'frame'

    if fname == '':
        fname = png_prefix + str(frameno).rjust(4,'0')
        if isinstance(figno,int):
            fname = fname + 'fig%s' % figno
    splitfname = os.path.splitext(fname)
    if splitfname[1] not in ('.png','.emf','.eps','.pdf'):
        fname = splitfname[0] + '.%s' % format
    if figno=='':
        figno = 1
    plt.figure(figno)
    if plotdir != '.':
        fname = os.path.join(plotdir,fname)
    if verbose:  print('    Saving plot to file ', fname)
    if kml_fig:
        # from webpage : https://robotics.usc.edu/~ampereir/wordpress/?p=626
        # This was added so that in KML, axes, tick labels, etc do not get printed.
        fig = plt.gcf()
        fig.patch.set_alpha(0)
        a = fig.gca()
        a.set_position([0.,0.,1.0,1.0])
        a.set_frame_on(False)

        # These have to be turned off explicitly so the plot doesn't have any padding.
        a.set_xticks([])
        a.set_yticks([])

        plt.axis('off')

        if kml_figsize is not None:
            fig.set_size_inches(kml_figsize[0],kml_figsize[1])
        a.set_frame_on(False)
        plt.subplots_adjust(top = 1, bottom = 0, right = 1, left = 0, 
                        hspace = 0, wspace = 0)
        plt.margins(0,0)
        plt.savefig(fname, transparent=True, bbox_inches='tight',dpi=kml_dpi)
    else:
        plt.savefig(fname, bbox_inches=bbox_inches)

    if close_fig:
        # to avoid running out of memory when making many plots
        plt.close(figno)


#======================================================================
def printframes(plotdata=None, verbose=True):
#======================================================================

    """
    Deprecated: use plotpages.plotclaw_driver instead to get gauges as well.
      - RJL, 1/1/10

    Produce a set of png files for all the figures specified by plotdata.
    Also produce a set of html files for viewing the figures and navigating
    between them.  These will all be in directorey plotdata.plotdir.

    The ClawPlotData object plotdata will be initialized by a call to
    function setplot unless plotdata.setplot=False.

    If plotdata.setplot=True then it is assumed that the current directory
    contains a module setplot.py that defines this function.

    If plotdata.setplot is a string then it is assumed this is the name of
    a module to import that contains the function setplot.

    If plotdata.setplot is a function then this function will be used.
    """
    from clawpack.visclaw import plotpages

    if 'matplotlib' not in sys.modules:
        print('*** Error: matplotlib not found, no plots will be done')
        return plotdata

    if not isinstance(plotdata,ClawPlotData):
        print('*** Error, plotdata must be an object of type ClawPlotData')
        return plotdata

    plotdata._mode = 'printframes'

    plotdata = call_setplot(plotdata.setplot, plotdata)

    try:
        plotdata.rundir = os.path.abspath(plotdata.rundir)
        plotdata.outdir = os.path.abspath(plotdata.outdir)
        plotdata.plotdir = os.path.abspath(plotdata.plotdir)

        framenos = plotdata.print_framenos # frames to plot
        fignos = plotdata.print_fignos     # figures to plot at each frame
        fignames = {}                      # names to use in html files

        rundir = plotdata.rundir       # directory containing *.data files
        outdir = plotdata.outdir       # directory containing fort.* files
        plotdir = plotdata.plotdir     # where to put png and html files
        overwrite = plotdata.overwrite # ok to overwrite?
        msgfile = plotdata.msgfile     # where to write error messages
    except AttributeError:
        print('*** Error in printframes: plotdata missing attribute')
        print('  *** plotdata = ',plotdata)
        return plotdata

    if fignos == 'all':
        fignos = plotdata._fignos
        #for (figname,plotfigure) in plotdata.plotfigure_dict.iteritems():
        #    fignos.append(plotfigure.figno)


    # filter out the fignos that will be empty, i.e.  plotfigure._show=False.
    plotdata = set_show(plotdata)
    fignos_to_show = []
    for figname in plotdata._fignames:
        figno = plotdata.plotfigure_dict[figname].figno
        if (figno in fignos) and plotdata.plotfigure_dict[figname]._show:
            fignos_to_show.append(figno)
    fignos = fignos_to_show


    rootdir = os.getcwd()

    # annoying fix needed when EPD is used for plotting under cygwin:
    cw_path = r'C:\cygwin'
    if rootdir[0:9] == cw_path and outdir[0:9] != cw_path:
        outdir = cw_path + outdir
        plotdata.outdir = outdir
    if rootdir[0:9] == cw_path and rundir[0:9] != cw_path:
        rundir = cw_path + rundir
        plotdata.rundir = rundir
    if rootdir[0:9] == cw_path and plotdir[0:9] != cw_path:
        plotdir = cw_path + plotdir
        plotdata.plotdir = plotdir

    try:
        os.chdir(rundir)
    except OSError:
        print('*** Error: cannot move to run directory ',rundir)
        print('rootdir = ',rootdir)
        return plotdata


    if msgfile != '':
        sys.stdout = open(msgfile, 'w')
        sys.stderr = sys.stdout


    try:
        plotpages.cd_plotdir(plotdata)
    except:
        print("*** Error, aborting plotframes")
        return plotdata


    framefiles = glob.glob(os.path.join(plotdir,'frame*.png')) + \
                    glob.glob(os.path.join(plotdir,'frame*.html'))
    if overwrite:
        # remove any old versions:
        for file in framefiles:
            os.remove(file)
    else:
        if len(framefiles) > 1:
            print("*** Remove frame*.png and frame*.html and try again,")
            print("  or use overwrite=True in call to printframes")
            return plotdata


    # Create each of the figures
    #---------------------------

    try:
        os.chdir(outdir)
    except OSError:
        print('*** Error printframes: cannot move to outdir = ',outdir)
        return plotdata


    fortfile = {}
    frametimes = {}

    for file in glob.glob('fort.q*'):
        frameno = int(file[6:])
        fortfile[frameno] = file

    if len(fortfile) == 0:
        print('*** No fort.q files found in directory ', os.getcwd())
        return plotdata

    # Discard frames that are not from latest run, based on
    # file modification time:
    framenos = only_most_recent(framenos, plotdata.outdir)

    numframes = len(framenos)

    print("Will plot %i frames numbered:" % numframes, framenos)
    print('Will make %i figure(s) for each frame, numbered: ' % len(fignos),\
          fignos)

    fignames = {}
    for figname in plotdata._fignames:
        figno = plotdata.plotfigure_dict[figname].figno
        fignames[figno] = figname



    # Make png files by calling plotframe:
    # ------------------------------------

    for frameno in framenos:
        #plotframe(frameno, plotdata, verbose)
        frametimes[frameno] = plotdata.getframe(frameno, plotdata.outdir).t
        #print 'Frame %i at time t = %s' % (frameno, frametimes[frameno])

    plotdata.timeframes_framenos = framenos
    plotdata.timeframes_frametimes = frametimes
    plotdata.timeframes_fignos = fignos
    plotdata.timeframes_fignames = fignames

    os.chdir(plotdir)

    if plotdata.html:
        plotpages.timeframes2html(plotdata)

    if not plotdata.printfigs:
        print("Using previously printed figure files")
    else:
        print("Now making png files for all figures...")
        for frameno in framenos:
            plotframe(frameno, plotdata, verbose)
            #frametimes[frameno] = plotdata.framesoln_dict[frameno].t
            print('Frame %i at time t = %s' % (frameno, frametimes[frameno]))

    if plotdata.latex:
        plotpages.timeframes2latex(plotdata)


    # Movie:
    #-------

    if plotdata.gif_movie:
        print('Making gif movies.  This may take some time....')
        for figno in fignos:
            try:
                os.system('convert -delay 20 frame*fig%s.png moviefig%s.gif' \
                   % (figno,figno))
                print('    Created moviefig%s.gif' % figno)
            except:
                print('*** Error creating moviefig%s.gif' % figno)

    os.chdir(rootdir)

    # print out pointers to html index page:
    path_to_html_index = os.path.join(os.path.abspath(plotdata.plotdir), \
                               plotdata.html_index_fname)
    plotpages.print_html_pointers(path_to_html_index)

    # reset stdout for future print statements
    sys.stdout = sys.__stdout__

    return plotdata
    # end of printframes


#------------------------------------------------------------------
def var_limits(plotdata,vars,padding=0.1):
#------------------------------------------------------------------
    """
    Determine range of values encountered in data for all frames
    in the most recent computation in order to determine appropriate
    y-axis limits for plots.

    vars is a list of variables to return the limits for.
    Each element of the list can be a number, indicating the
    corresponding element of the q vector, or a function that should
    be applied to q to obtain the value of interest.  In other words,
    any valid plot_var attribute of a ClawPlotItem object.

    If vars=='all', use vars=[0,1,...] over all components of q.
    *** not implemented yet ***

    Returns: varmin, varmax, varlim,
    each is a dictionary with keys consisting by the elements of vars.

    For each var in vars,
      varmin[var] is the minimum of var over all frames,
      varmax[var] is the maximum of var over all frames,
      varlim[var] is of the form [v1, v2] suitable to use as the
            ylimits attritube of an object of class ClawPlotAxes with
                v1 = vmin - padding*(vmax-vmin)
                v2 = vmax + padding*(vmax-vmin)
            to give some space above and below the min and max.

    """

    varlim = {}
    vmin,vmax = var_minmax(plotdata,'all',vars)
    varmin = {}
    varmax = {}
    for var in vars:
        varmin[var] = vmin[var]['all']   # min over all frames
        varmax[var] = vmax[var]['all']   # max over all frames
        v1 = varmin[var] - padding*(varmax[var]-varmin[var])
        v2 = varmax[var] + padding*(varmax[var]-varmin[var])
        varlim[var] = [v1,v2]

    return varmin,varmax,varlim


#------------------------------------------------------------------
def var_minmax(plotdata,framenos,vars):
#------------------------------------------------------------------
    """
    Determine range of values encountered in data for all frames
    in the list framenos.  If framenos='all', search over all frames
    in the most recent computation.

    vars is a list of variables to return the min and max for.
    Each element of the list can be a number, indicating the
    corresponding element of the q vector, or a function that should
    be applied to q to obtain the value of interest.  In other words,
    any valid plot_var attribute of a ClawPlotItem object.

    If vars=='all', use vars=[0,1,...] over all components of q.
    *** not implemented yet ***

    Returns (varmin, varmax) where varmin and varmax are each
    dictionaries with keys given by the variables.

    For each var in vars, varmin[var] and varmax[var] are dictionaries, with
    keys given by frame numbers in framenos.

    Also varmin[var]['all'] and varmax[var]['all'] are the min and max of
    variable var over all frames considered.

    For example if vars = [0,'machnumber'] then
       varmin[0][3] is the minimum of q[0] (the first component of the
           solution vector q) over all patches in frame 3.
       varmin['machnumber']['all'] is the minimum of machnumber
           over all patches in all frames.

    """

    framenos = only_most_recent(framenos, plotdata.outdir)
    if len(framenos) == 0:
        print('*** No frames found in var_minmax!')

    print('Determining min and max of variables: ',vars)
    print('   over frames: ',framenos)

    varmin = {}
    varmax = {}
    if vars=='all':
        print("*** Error in var_minmax: vars == 'all' is not implemented yet")
        return (varmin,varmax)
    for var in vars:
        varmin[var] = {'all': np.inf}
        varmax[var] = {'all': -np.inf}
        for frameno in framenos:
            varmin[var][frameno] = np.inf
            varmax[var][frameno] = -np.inf

    for frameno in framenos:
        solution = plotdata.getframe(frameno, plotdata.outdir)
        num_dim = solution.num_dim

        for state in solution.states:
            patch = state.patch
            for var in vars:
                if isinstance(var,int):
                    if num_dim == 1:
                        qvar = state.q[var,:]
                    elif num_dim == 2:
                        qvar = state.q[var,:,:]
                    elif num_dim == 3:
                        qvar = state.q[var,:,:,:]
                else:
                    t = solution.t
                    #patch.compute_physical_coordinates()
                    if num_dim == 1:
                        X_center = patch.grid.p_centers[0]
                        qvar = var(state.q, X_center, t)
                    elif num_dim == 2:
                        X_center, Y_center = patch.grid.p_centers
                        qvar = var(state.q, X_center, \
                                   Y_center, t)
                    elif num_dim == 3:
                        X_center, Y_center, Z_center = patch.grid.p_centers
                        qvar = var(state.q, X_center, \
                                   Y_center, Z_center, t)
                varmin[var][frameno] = min(varmin[var][frameno], qvar.min())
                varmax[var][frameno] = max(varmax[var][frameno], qvar.max())
                varmin[var]['all'] = min(varmin[var]['all'], \
                                         varmin[var][frameno])
                varmax[var]['all'] = max(varmax[var]['all'], \
                                         varmax[var][frameno])
    return (varmin, varmax)



#------------------------------------------------------------------
def only_most_recent(framenos,outdir='.',prefix='fort',verbose=True):
#------------------------------------------------------------------

    """
    Filter the list framenos of frame numbers and return a new
    list that only contains the ones from the most recent run.
    This is determined by comparing modification times of
    fort.q files.

    Returns the filtered list.
    """

    startdir = os.getcwd()
    if outdir != '.':
        try:
            os.chdir(outdir)
        except OSError:
            print("*** Could not chdir to ", outdir)
            return framenos

    fortfile = {}
    for file in glob.glob(prefix + '.q*'):
        frameno = int(file[-4:])
        fortfile[frameno] = file

    #DK: In PetClaw, we don't output fort.q* files.  Instead count the
    #claw.pkl* files.
    if len(fortfile) == 0:
        for file in glob.glob('claw.pkl*'):
            frameno = int(file[8:])
            fortfile[frameno] = file

    if len(fortfile) == 0:
        print('*** No fort.q or claw.pkl files found in directory ', os.getcwd())
        framenos = []
        return framenos

    # Figure out which files are from latest run:
    numframes = 0
    mtime = 0
    framekeys = list(fortfile.keys())
    framekeys.sort()
    for frameno in framekeys:
        mtimeprev = mtime
        mtime = os.path.getmtime(fortfile[frameno])
        # sometimes later fort files are closed a few seconds after
        # earlier ones, so include a possible delaytime:
        delaytime = 5  # seconds
        if mtime < mtimeprev-delaytime:
            break
        numframes = numframes + 1

    newframes = framekeys[:numframes]
    if (numframes < len(framekeys)) & verbose:
        print('*** Frames %s and above appear to be from an old run' \
                       % framekeys[numframes])
        print('***    and will be ignored.')
        time.sleep(2)

    #print 'framenos = ',framenos
    if framenos == 'all':
        framenos = newframes
    else:
        # compute intersection of framenos and newframes:
        framenos = list(set(framenos).intersection(set(newframes)))
    framenos.sort()
    os.chdir(startdir)
    return framenos

#------------------------------------------------------------------------
def call_setplot(setplot, plotdata, verbose=True):
#------------------------------------------------------------------
    """
    Try to apply setplot to plotdata and return the result.
    If setplot is False or None, return plotdata unchanged.
    If setplot is True, setplot function is in setplot.py.
    If setplot is a string, setplot function is in module named by string.
    Otherwise assume setplot is a function.
    """
    # This is a bit of a hack to make sure that we still handle the
    # setplot == None case, we may want to deprecate this and require
    # an argument here
    if setplot is None:
        setplot = False

    # Figure out what to do with the setplot passed in
    if not isinstance(setplot,types.FunctionType):
        # We need to parse the names given and import the module and function
        if isinstance(setplot,bool):
            if setplot:
                # Assume that we should import setplot.py from current directory
                setplot_module_dir = os.getcwd()
                setplot_module_name = "setplot"
            else:
                if verbose:
                    print("*** WARNING:  No setplot specified!", file=sys.stderr)
                return plotdata
        elif isinstance(setplot,str):
            # Assume this is the path to the required setplot module
            path = os.path.abspath(os.path.expandvars(os.path.expanduser(setplot)))
            setplot_module_dir = os.path.dirname(path)
            setplot_module_name = os.path.splitext(os.path.basename(setplot))[0]
        else:
            raise Exception("Invalid setplot module specification.")

        if verbose:
            print("Importing %s.setplot from %s." % (setplot_module_name,setplot_module_dir))

        # Attempt to import whatever was handed to us and parsed above
        try:
            sys.path.insert(0,setplot_module_dir)
            setplot_module = __import__(setplot_module_name)
            reload(setplot_module)
            setplot = setplot_module.setplot
            if not isinstance(setplot,types.FunctionType):
                raise ImportError("Failed importing %s.setplot" % setplot_module_name)
        except ImportError as e:
            print("WARNING: Failed to import %s.setplot from %s." \
                                        % (setplot_module_name,setplot_module_dir), file=sys.stderr)
            print("Exception and traceback:", file=sys.stderr)
            print("="*80, file=sys.stderr)
            print("\t%s\n\n" % e, file=sys.stderr)
            traceback.print_exc(file=sys.stderr)
            print("="*80, file=sys.stderr)

            # Not sure if this is the right place to be doing this, maybe
            # this should go higher up than this routine so it can check all
            # of the same attributes?
            print("Would you like to use clawpack.visclaw.setplot_default() instead [Y/n]?")
            use_default = input()
            if (use_default == "") or ("Y" in use_default.capitalize()):
                setplot = setplot_default.setplot
            else:
                sys.exit(1)

            plotdata = setplot(plotdata)
            return plotdata
        except:
            raise
    else:
        # setplot was handed to us as a function, we don't have to do anything!
        pass

    # Execute the setplot function
    try:
        plotdata = setplot(plotdata)
        if plotdata is None:
            # Simple warning if we did not get anything back from plotdata
            print('*** Did you forget the "return plotdata" statement?')
            raise Exception("Plotdata was not set by given setplot function!")
        if verbose:
            print('Executed setplot successfully')
    except:
        print('*** Error in call_setplot: Problem executing function setplot', file=sys.stderr)
        raise

    return plotdata


#------------------------------------------------------------------
def clawpack_header():
#------------------------------------------------------------------
    plt.axes([.3, .8, .98, 1.])
    plt.text(.1,.13,'Clawpack Plots',fontsize=30,color='brown')
    plt.axis('off')

#------------------------------------------------------------------
def errors_2d_vs_1d(solution,reference,var_2d,var_1d,map_2d_to_1d):
#------------------------------------------------------------------
    """
    Input:
      solution: object of class ClawSolution with 2d computed solution

      reference: object of class ClawSolution with 1d reference solution

      var_2d: variable to compare from solution
      if integer, compare q[var_2d,:,:]
      if function, apply to q to obtain the variable

      var_1d: variable to compare from reference
      if integer, compare to q[var_2d,:]
      if function, apply to q to obtain the variable

      map_2d_to_1d: function mapping 2d data to 1d for comparison::

          xs, qs = map_2d_to_1d(xcenter, ycenter, q)
    """

    t = solution.t
    xs = {}
    qs = {}
    qint = {}

    errmax = 0.
    for stateno,state in enumerate(solution.states):
        patch = state.patch

        X_center, Y_center = patch.grid.p_centers
        X_edge, Y_edge = patch.grid.p_centers

        if isinstance(var_2d, int):
            q = state.q[var_2d,:,:]
        else:
            try:
                q = var_2d(state.q, X_center, Y_center, t)
            except:
                print('*** Error applying function plot_var = ',plot_var)
                traceback.print_exc()
                return

        xs1, qs1 = map_2d_to_1d(q, X_center, Y_center, t)

        if hasattr(reference,'states'):
            if len(reference.states) > 1:
                print('*** Warning in errors_2d_vs_1d: reference solution')
                print('*** has more than one patch -- only using patch[0]')
            refstate = reference.states[0]
        else:
            refstate = reference   # assume this contains true solution or
                                  # something set separately rather than
                                  # a framesoln

        xref = patch.grid.p_centers[0]
        if isinstance(var_1d, int):
            qref = refstate.q[var_1d,:].T
        else:
            try:
                qref = var_1d(refstate.q, xref, t)
            except:
                print('*** Error applying function var_1d')
                return

        qint1 = np.interp(xs1, xref, qref)

        xs[stateno] = xs1
        qs[stateno] = qs1
        qint[stateno] = qint1
        errabs = abs(qs1 - qint1)
        errmax = max(errmax, errabs.max())

    return errmax, xs, qs, qint



#------------------------------------------------------------------
def set_show(plotdata):
#------------------------------------------------------------------
    """
    Determine which figures and axes should be shown.
    plotaxes._show should be true only if plotaxes.show and at least one
    item is showing or if the axes have attribute type=='empty', in which
    case something may be plotting in an afteraxes command, for example.

    plotfigure._show should be true only if plotfigure.show and at least
    one axes is showing.
    """

    for figname in plotdata._fignames:
        plotfigure = plotdata.plotfigure_dict[figname]
        plotfigure._show = False
        if plotfigure.show:
            # Loop through all axes to make sure at least some item is showing
            for plotaxes in plotfigure.plotaxes_dict.values():
                plotaxes._show = False
                if plotaxes.show:
                    # Loop through plotitems checking each item to see if it
                    # should be shown
                    for plotitem in plotaxes.plotitem_dict.values():
                        plotitem._show = plotitem.show
                        if plotitem.show:
                            plotaxes._show = True
                            plotfigure._show = True
                    # Check to see if the axes are supposed to be empty or
                    # something may be in the afteraxes function
                    if not plotaxes._show:
                        if plotaxes.afteraxes is not None or plotaxes.type == 'empty':
                            plotaxes._show = True
                            plotfigure._show = True

    return plotdata



#!/usr/bin/env python

"""Module contains class definitions related to dealing with gauge data"""

import os
import sys
import numpy

# See if pandas is available
try:
    import pandas
    pandas_available = True
except ImportError as e:
    pandas_available = False


class GaugeSolution(object):
    r"""Object representing a numerical gauge station

    The `GaugeSolution` class provides a generic means of reading and
    interacting with gauge data output from a run.  A header describing the
    contents of the data is read in and

    :Initialization:

        There are a few ways to create a `GaugeSolution`:
          1. Provide a `gauge_id` and a `path` to find that gauge.  This will
             initialize a gauge with the data at `path` for the gauge with gauge
             id `gauge_id`.  Note that `path` will default to the current
             working directory provided by `os.getcwd()` if `path` is not
             provided.
          2. Create an empty gauge by providing neither `gauge_id` or `path`.

        Additionally the parameter `use_pandas` can be provided which determines
        how the gauge's data is stored.  By default this is `False` and the
        `GaugeSolution` will use a `numpy` ndarray.
    """


    def __init__(self, gauge_id=None, path=None, use_pandas=False):
        r"""GaugeSolution Initiatlization Routine

        See :class:`GaugeSolution` for more info.
        """

        # Gauge descriptors
        self.id = None
        r"""(int) - Gauge id of this gauge.  Also known as gauge number"""
        self.location = None
        r"""(tuple) - Location of the gauge recorded in the header"""

        # Gauge data
        self.level = None
        r"""(ndarray(:) - int) - The level the given observation used"""
        self.t = None
        r"""(ndarray(:) - float) - The time the given observation was made at"""
        self.q = None
        r"""(ndarray(*, :) - float) - Observed data"""
        self.aux = None
        r"""(ndarray(*, :) - float) - Auxiliary data"""
        self.gtype = 'stationary'
        r"""(str) - 'stationary' or 'lagrangian'"""
        self.particle_path = None
        r"""(ndarray(:,3) - for lagrangian gauge columns are t,x,y,
            for stationary gauge, single row with [t0,x,y]
            Only implemented in 2d so far."""

        # Read in gauge data from file
        if gauge_id is not None:
            if path is None:
                path = os.getcwd()
            self.read(gauge_id, path, use_pandas)


    def read(self, gauge_id, path=None, use_pandas=False):
        r"""Read the gauge file at path into this object

        Read in the gauge with gauge id `gauge_id` located at `path`.  

        If `use_pandas` is `True` then `q` will be a Pandas frame.
            (not yet implemented)

        New in v5.9.0: If gauge00000N.txt file contains only a header,
        read data from corresponding .bin file.

        :Input:
         - *gauge_id* - (int) Gauge id to be read
         - *path* - (path) Path to directory containing the gauge data.
           Defaults to `path = os.getcwd()`.
         - *use_pandas* - (bool) If True then `q` will be a Pandas frame,
           otherwise it will be a `numpy.ndarary`.  Default to `False`.

        :Output:
         None
        """

        # Construct path to gauge
        if path is None:
            path = os.getcwd()

        # First read header from .txt file:
        gauge_file_name = "gauge%s.txt" % str(gauge_id).zfill(5)
        gauge_path = os.path.join(path, gauge_file_name)
        if not os.path.isfile(gauge_path):
            print('Did not find %s' % gauge_path)
            file_format = 'binary'

        with open(gauge_path, 'r') as gauge_file:
            # First line
            data = gauge_file.readline().split()
            self.id = int(data[2])
            if len(data) == 8:
                # 1d
                self.location = (float(data[4]),)
                num_eqn = int(data[7])
            if len(data) == 9:
                # 2d
                self.location = (float(data[4]), float(data[5]))
                num_eqn = int(data[8])
            elif len(data) == 10:
                # 3d
                self.location = (float(data[4]), float(data[5]), float(data[6]))
                num_eqn = int(data[9])

            line = gauge_file.readline()
            if 'lagrangian' in line.lower():
                # Lagrangian gauge (particle)
                self.gtype = 'lagrangian'
                line = gauge_file.readline() # third line of header
            elif 'stationary' in line.lower():
                # Standard stationary gauge
                self.gtype = 'stationary'
                line = gauge_file.readline() # third line of header
            else:
                # backward compatibility
                self.gtype = 'stationary'

            #data = line.split()
            #nvals = 1 + len(data)-6  # should agree with num_eqn
            
            # Check to see if there is also data in the .txt file,
            # otherwise perhaps it's in a binary .bin file.
            
            line = gauge_file.readline()
            if len(line) > 0:
                if 'binary32' in line:
                    file_format = 'binary32'
                elif 'binary' in line:
                    file_format = 'binary'  # also allows 'binary64'
                else:
                    # if 'ascii' in line, or no file_format line (data follows)
                    # (for backward compatibility)
                    file_format = 'ascii'
            else:
                # if file_format line missing and no data lines, try binary:
                file_format = 'binary'

        # Check to see if the gauge file name ID and that inside of the gauge
        # file are the same
        gauge_file_name = os.path.basename(path)
        if self.id != gauge_id:
            raise ValueError("Gauge ID requested does not match ID inside ",
                             "file!")

        # Read gauge data

        if use_pandas:
            raise NotImplementedError("Pandas data backend not implemented yet.")
            if not pandas_available:
                raise ImportError("Pandas not available.")


        if file_format == 'ascii':
            # data follows header in .txt file:
            data = numpy.loadtxt(gauge_path, comments="#")
            if data.ndim == 1:
                # only one line in gauge file, expand to 2d array
                data = data.reshape((1,len(data)))

        if file_format[:6] == 'binary':
            # data is in separate .bin file:
            gauge_file_name = "gauge%s.bin" % str(gauge_id).zfill(5)
            gauge_path = os.path.join(path, gauge_file_name)
            if not os.path.isfile(gauge_path):
                msg = 'No data in .txt file and did not find ' \
                         + '\n   binary file %s' %  gauge_path
                import warnings
                warnings.warn(msg)
                return

            if file_format in ['binary','binary64']:
                data = numpy.fromfile(gauge_path, dtype=numpy.float64)
            elif file_format == 'binary32':
                data = numpy.fromfile(gauge_path, dtype=numpy.float32)

            # assume rows are: level, t, q[0:num_eqn]
            nrows = 2 + num_eqn
            assert numpy.mod(len(data),nrows) == 0, \
                  '*** unexpected number of values in gauge file' \
                  + '\n*** expected nrows = %i rows'  % nrows
            ncols = int(len(data)/nrows)
            data = data.reshape((nrows,ncols), order='F').T

        self.level = data[:, 0].astype(numpy.int64)
        self.t = data[:, 1]
        self.q = data[:, 2:].transpose()

    
        if num_eqn != self.q.shape[0]:
            raise ValueError("Number of fields in gauge file does not match",
                             "recorded number in header.")

        if len(self.location) == 2:
            # lagrangian gauges only implemented in 2d so far
            if self.gtype == 'lagrangian':
                # for lagrangian gauges x,y paths are stored in q in place of u,v
                # note: only implemented in 2d geoclaw for now!
                self.particle_path = numpy.vstack((self.t, self.q[1,:], 
                                                  self.q[2,:])).T
            else:
                # set the lagrangian path to a single fixed location at t=t0:
                self.particle_path = numpy.array([[self.t[0], self.location[0], 
                                             self.location[1]]])



    def write(self, path=None, format="%+.15e"):
        r"""Write the data from this gauge to a file in `path`

        :Input:
         - *path* (path) Path to write the gauge file to.  Defaults to
           `path = os.getcwd()`.
         - *format* (str) Format string used for the field values.

        :Output:
         None
        """

        if path is None:
            path = os.getcwd()

        if not self.is_valid():
            raise ValueError("Gauge is not initialized properly.")

        gauge_file_name = "gauge%s.txt" % str(self.id).zfill(5)
        with open(os.path.join(path, gauge_file_name), "w") as gauge_file:

            gauge_file.write("# gauge_id= %s location=( %s %s ) num_eqn= %s\n" %
                 (self.id, self.location[0], self.location[1], self.q.shape[0]))
            gauge_file.write("# Columns: level time q(1 ... num_eqn)\n")

            # print(self.q.shape)
            for i in range(self.q.shape[1]):
                gauge_file.write("%02i %+.15e " % (self.level[i], self.t[i]))
                gauge_file.write(" ".join([format % value for value in self.q[:, i]]))
                gauge_file.write("\n")


    def is_valid(self):
        r"""Check to see if the gauge data has all been set.

        Simply checks to see if all attributes are not None.

        :Output:
         - (bool) Whether the gauge data has been set.
         """

        if ((self.id is not None) and (self.location is not None) and
            (self.level is not None) and (self.t is not None) and
            (self.q is not None)):

            return True

        return False


    def __repr__(self):
        if self.is_valid():
            output = "%4i" % self.id
            for j in range(len(self.location)):
                output = " ".join((output,"%17.10e" % self.location[j]))
            output = " ".join((output,"%13.6e" % self.t[0]))
            output = " ".join((output,"%13.6e\n" % self.t[-1]))
        else:
            output = None

        return output


    def __str__(self):
        return ("Gauge %s: location = %s, t = [%s, %s], %s" %
                                    (self.id, self.location,
                                     self.t[0], self.t[-1], self.gtype))


# ==============================
#  Utility Functions for Gauges
# ==============================
def compare_gauges(paths, gauge_id, fields='all'):
    r"""Make plots comparing gauges in different directories

    :Input:
     - *paths* (list) List of paths of length 2 pointing to the directories that
       the gauge files are to be taken from.
     - *gauge_id* (int) Gauge id to compare.
     - *fields* (int or list) Fields to be plotted.  If fields == 'all' then all
       available fields will be plotted.  Default is 'all'.

    :Output:
     - (matplotlib.figure.Figure) Figure object created by comparison
    """

    import matplotlib.pyplot as plt

    if len(paths) != 2:
        raise ValueError("Provide two paths to gauge files for comparison.")

    gauges = []
    for path in paths:
        gauges.append(GaugeSolution(path=path, gauge_id=gauge_id))

    if isinstance(fields, str):
        if fields.lower() == 'all':
            fields = range(gauges[0].q.shape[0])

    fig = plt.figure()
    fig.suptitle("Gauge %s" % gauges[0].id)
    for (i, n) in enumerate(fields):
        axes = fig.add_subplot(len(fields), 2, 2 * i + 1, )
        axes.plot(gauges[0].t, gauges[0].q[n, :], 'ko', label="%s" % paths[0])
        axes.plot(gauges[1].t, gauges[1].q[n, :], 'rx', label="%s" % paths[1])
        axes.set_xlabel("t")
        axes.set_ylabel("q[%s, :]" % n)
        axes.legend()

        axes = fig.add_subplot(len(fields), 2, 2 * i + 2)
        axes.plot(gauges[0].t, 
                  numpy.abs(gauges[0].q[n, :] - gauges[1].q[n, :]), 'r')
        axes.set_xlabel("t")
        axes.set_ylabel("$|q_{old}[%s, :] - q_{new}[%s, :]|$" % (n, n))

    return fig



# =============================================================
#  Utility Functions to Help Transition to the New Gauge Files
# =============================================================
def convert_gauges(path, output_path=None):
    r"""Convert gauge output data from fort.gauge file to new format

    This is provided for data files that were in the old format and will split
    up the data in the old `fort.gauge` files into separate files for each
    gauge.

    :Input:
     - *path* - (path) Path to old file.
     - *output_path* - (path) Path to directory the new gauge files will be
       placed.  Defaults to `output_path = os.getcwd()`.
    """

    if output_path is None:
        output_path = os.getcwd()

    # Load old data
    data = numpy.loadtxt(path)
    old_ids = numpy.asarray(data[:, 0], dtype=int)
    unique_ids = numpy.asarray(list(set(old_ids)))

    # Create new gauges and compare
    for gauge_id in unique_ids:
        gauge_indices = numpy.nonzero(old_ids == gauge_id)[0]
        new_gauge = GaugeSolution()
        new_gauge.id = gauge_id
        new_gauge.location = (numpy.nan, numpy.nan)
        new_gauge.level = numpy.asarray(data[gauge_indices, 1], dtype=int)
        new_gauge.t = data[gauge_indices, 2]
        new_gauge.q = data[gauge_indices, 3:].transpose()
        new_gauge.write(output_path)

        # Test gauge data
        if not compare_gauges(path, output_path, gauge_id):
            print("WARNING:  New gauge data does not match old gauge data.")
            print("          gauge id = %s" % gauge_id)



def compare_old_gauges(old_path, new_path, gauge_id, plot=False, abs_tol=1e-14,
                                                     rel_tol=0.0, 
                                                     verbose=False):
    r"""Compare old gauge data at `path` to new gauge data at same path

    Provided as a quick check to see if the function `convert_gauges` has
    correctly converted the data located in `fort.gauge`.  Really meant as a
    simple check but can be used to compare new test data to old.

    :Input:
     - *old_path* - (path) Path to old file.
     - *new_path* - (path) Path to new gauge files (directory).
     - *gauge_id* - (int) Gauge id to compare.
     - *plot* - (bool) Whether to plot the differences between the gauges.
     - *old_file_name* - (string) Name of the old gauge file, defaults to
       `old_file_name = "fort.gauge"`
     - *abs_tol* - (float) Absolute tolerance used to make the comparison.  See
       `numpy.allclose` for more info.
     - *rel_tol* - (float) Relative tolerance used to make the comparison.  See
       `numpy.allclose` for more info.

    :Output:
     - (bool) Whether the gauges agreed to double precision.  Uses
       `numpy.testing.assert_allequal` to check with the `abs_tol` and `rel_tol`
       specified above.
    """

    # Load old gauge data
    data = numpy.loadtxt(old_path)
    old_ids = numpy.asarray(data[:, 0], dtype=int)
    gauge_indices = numpy.nonzero(old_ids == gauge_id)[0]
    q = data[gauge_indices, 3:]

    # Load new data
    gauge = GaugeSolution(gauge_id, new_path)

    # Turn these into assertions or logicals
    if verbose:
        print("Comparison of gauge %s:" % gauge_id)
        print(r"           ||\Delta q||_2 = ",
                              numpy.linalg.norm(q - gauge.q.transpose(), ord=2))
        print(r"  arg(||\Delta q||_\infty = ",
                              numpy.argmax(q - gauge.q.transpose()))

    if plot:
        import matplotlib.pyplot as plt
        fig = plt.figure()
        for i in range(gauge.q.shape[0]):
            axes = fig.add_subplot(1, 3, i + 1)
            axes.plot(q[:, i] - gauge.q[i, :])
            axes.set_title("q[%s, :] comparison" % i)
            axes.set_xlabel("t (s)")
            axes.set_ylabel("q(%s, :)" % i)

    return numpy.allclose(q, gauge.q.transpose(), rtol=rel_tol, atol=abs_tol)


def check_old_gauge_data(path, gauge_id, new_gauge_path="./regression_data"):
    """Compare old gauge data to new gauge format

    Function is meant to check discrepancies between versions of the gauge
    files.  Note that this function also directly prints out some info.

    :Input:
     - *path* (string) - Path to old gauge data file
     - *gauge_id* (int) - Gauge id to compare
     - *new_gauge_path* (path) - Path to directory containing new gauge files, 
       defaults to './regression_data'.

    :Output:
     - (figure) Returns a matplotlib figure object plotting the differences in 
       time.
    """

    import matplotlib.pyplot as plt

    # Load old gauge data
    data = numpy.loadtxt(path)
    old_ids = numpy.asarray(data[:, 0], dtype=int)
    gauge_indices = numpy.nonzero(old_ids == gauge_id)[0]
    q = data[gauge_indices, 3:]

    # Load new data
    gauge = GaugeSolution(gauge_id, new_gauge_path)

    print(numpy.linalg.norm(q - gauge.q.transpose(), ord=2))
    print(numpy.argmax(q - gauge.q.transpose()))

    fig = plt.figure()
    for i in range(gauge.q.shape[0]):
        axes = fig.add_subplot(1, gauge.q.shape[0], i + 1)
        axes.plot(q[:, i] - gauge.q[i, :])
        axes.set_title("q[%s, :] comparison" % i)

    return fig


if __name__ == "__main__":

    import matplotlib.pyplot as plt

    help_msg = \
"""gauges.py path1 path2 [gauge_id] [fields...]

Plots a comparison between the gauges at path1 and path2 with gauge_id and the 
fields specified.  Only one gauge_id can be specified at a time but a number of 
fields can be specified including 'all'.
"""

    fields = 'all'
    gauge_id = 1
    if len(sys.argv) < 3:
        print(help_msg)
        sys.exit(0)
    elif len(sys.argv) >= 3:
        paths = [str(sys.argv[1]), str(sys.argv[2])]
        if len(sys.argv) > 3:
            gauge_id = int(sys.argv[3])
            if len(sys.argv) > 4:
                if sys.argv[4].lower() == 'all':
                    fields = 'all'
                else:
                    fields = [int(field for field in sys.argv[4:])]


    fig = compare_gauges(paths, gauge_id, fields)
    plt.show()



#!/usr/bin/env python
# encoding: utf-8
r"""
Simple advection Riemann solvers

Basic advection Riemann solvers of the form (1d)

.. math::
    q_t + A q_x = 0.
    
:Authors:
    Kyle T. Mandli (2008-2-20): Initial version
"""
# ============================================================================
#      Copyright (C) 2008 Kyle T. Mandli <mandli@amath.washington.edu>
#
#  Distributed under the terms of the Berkeley Software Distribution (BSD) 
#  license
#                     http://www.opensource.org/licenses/
# ============================================================================

# Riemann solver constants
num_eqn = 1
num_waves = 1

import numpy as np

def advection_1D(q_l,q_r,aux_l,aux_r,problem_data):
    r"""Basic 1d advection riemann solver
    
    *problem_data* should contain -
     - *u* - (float) Determines advection speed
    
    See :ref:`pyclaw_rp` for more details.
    
    :Version: 1.0 (2008-2-20)
    """
    
    # Number of Riemann problems we are solving
    num_rp = q_l.shape[1]

    # Return values
    wave = np.empty( (num_eqn, num_waves, num_rp) )
    s = np.empty( (num_waves, num_rp) )
    amdq = np.zeros( (num_eqn, num_rp) )
    apdq = np.zeros( (num_eqn, num_rp) )
 
    wave[0,0,:] = q_r[0,:] - q_l[0,:]
    s[0,:] = problem_data['u']
    if problem_data['u'] > 0:
        apdq[0,:] = s[0,:] * wave[0,0,:]
    else:
        amdq[0,:] = s[0,:] * wave[0,0,:]

    return wave, s, amdq, apdq



#!/usr/bin/python

"""
fixdata
=========
    Provides functions for removing erroneous data from arrays as grids

    Contains:
        findbadindices
        fillbaddata


"""

import string
import re
import numpy
from numpy import *

from . import iotools

#==============================================================================
def findbadindices (Z,badvalue=inf,removenans=True):
    """
        remove nans or infs from an array
    """

    badind=[]
    for i in range(shape(Z)[0]) :
        for j in range(shape(Z)[1]):

            if Z[i,j]==inf:
                badind.append((i,j))
            elif Z[i,j]==badvalue:
                badind.append((i,j))

            if Z[i,j]!=Z[i,j] and removenans:
                badind.append((i,j))


    return badind

#===============================================================================
def fillbaddata (Z,badinds):

    """
    fill data in array Z, at indice tuples in list badinds
    by averaging surrounding good data.
    return new array.
    """
    
    m=shape(Z)[0]
    n=shape(Z)[1]

    for ind in badinds :
        i=ind[0]
        j=ind[1]
        r=0
        indbad=True
        while indbad and r < max(m,n):
            r=r+1 #radius of ball around badinds in inf-norm (square)
            irange=list(range(max(0,i-r),min(i+r+1,m)))
            jrange=list(range(max(0,j-r),min(j+r+1,n)))
            summands=0
            sum=0.
            for ii in irange:
                for jj in jrange:
                    ballind=(ii,jj)
                    if not ballind in badinds:
                        sum = sum + Z[ballind[0],ballind[1]]
                        summands=summands+1
            if summands >0 : 
                Z[ind[0],ind[1]] = sum/summands
                indbad=False

    return Z


#=====================================================================================
def filterdata (Z,filterinds,radius=1):
    """
    filter data in array z, at indice tuples in list filterinds
    by averaging surrounding data, ball with radius=radius in inf-norm
    acts as a low-band pass filter and removes oscillatory data
    """

    m=shape(Z)[0]
    n=shape(Z)[1]

    for ind in filterinds :
        i=ind[0]
        j=ind[1]
        r=radius

        irange=list(range(max(0,i-r),min(i+r+1,m)))
        jrange=list(range(max(0,j-r),min(j+r+1,n)))
        summands=0
        sum=0.
        for ii in irange:
            for jj in jrange:
                ballind=(ii,jj)
                sum = sum + Z[ballind[0],ballind[1]]
                summands=summands+1
        if summands >0 : 
            Z[ind[0],ind[1]] = sum/summands

    return Z







